//
//  Bitboard.h
//  fornax2_core
//
//  Created by Anders on 25.06.2018.
//  Copyright © 2018 Avividi. All rights reserved.
//

#ifndef bitboard_h
#define bitboard_h

#include <assert.h>
#include "definitions.h"
#include "square.h"

#define BITBOARD_SET(bits, position) ((bits) |= ((bitboard64)1U) << (position))
static inline void bitboard_set(bitboard64* bits, square position) {
  assert(SQUARE_INBOUNDS(position));
  *bits |= ((bitboard64)1U) << position;
}

#define BITBOARD_CLEAR(bits, position) ((bits) &= ~(((bitboard64)1U) << (position)))
static inline void bitboard_clear(bitboard64* bits, square position) {
  assert(SQUARE_INBOUNDS(position));
  *bits &= ~(((bitboard64)1U) << position);
}

#define BITBOARD_TEST(bits, position) (((bits) >> (position)) & 1U)
static inline bool bitboard_test(const bitboard64 bits, square position) {
  assert(SQUARE_INBOUNDS(position));
  return (bits >> position) & 1U;
}

#define BITBOARD_COUNT(bits) (__builtin_popcountll(bits))
static inline int bitboard_count(const bitboard64 bits) {
  return __builtin_popcountll(bits);
}


#define BITBOARD_FIND_TRAILING(bits) (__builtin_ffsll(bits) - 1U)
static inline square bitboard_findTrailing(const bitboard64 bits) {
  return __builtin_ffsll(bits) - 1U;
}

static inline square bitboard_popTrailing(bitboard64* bits) {
  //  if (*bits == 0) return -1;
  
  square position = BITBOARD_FIND_TRAILING(*bits);
  if (position == -1) return -1;
  assert(position == -1 || SQUARE_INBOUNDS(position));
  BITBOARD_CLEAR(*bits, position);
  return position;
}

static inline square bitboard_findLeading(const bitboard64 bits) {
  return bits == 0 ? -1 : 63 - __builtin_clzll(bits);
}

static inline square bitboard_popLeading(bitboard64* bits) {
  square position = bitboard_findLeading(*bits);
  if (position < 0) return -1;
  
  bitboard_clear(bits, position);
  return position;
}

#define BITBOARD_FLIP(bits) __builtin_bswap64(bits)
static inline bitboard64 bitboard_flip(bitboard64 bits) {
  return __builtin_bswap64(bits);
}

#define BITBOARD_SHIFT_NW(bits) ((bits << 7) & ~BB_FILE_H)
#define BITBOARD_SHIFT_N(bits) (bits << 8)
#define BITBOARD_SHIFT_NE(bits) ((bits << 9) & ~BB_FILE_A)
#define BITBOARD_SHIFT_E(bits) ((bits << 1) & ~BB_FILE_A)
#define BITBOARD_SHIFT_SE(bits) ((bits >> 7) & ~BB_FILE_A)
#define BITBOARD_SHIFT_S(bits) (bits >> 8)
#define BITBOARD_SHIFT_SW(bits) ((bits >> 9) & ~BB_FILE_H)
#define BITBOARD_SHIFT_W(bits) ((bits >> 1) & ~BB_FILE_H)

static inline bitboard64 bitboard_fill_NW(bitboard64 bits) {
  bits |= (bits << 7); bits |= (bits << 14); bits |= (bits << 28);
  return bits;
}

static inline bitboard64 bitboard_fill_N(bitboard64 bits) {
  bits |= (bits << 8); bits |= (bits << 16); bits |= (bits << 32);
  return bits;
}

static inline bitboard64 bitboard_fill_NE(bitboard64 bits) {
  bits |= (bits << 9); bits |= (bits << 18); bits |= (bits << 36);
  return bits;
}

static inline bitboard64 bitboard_fill_E(bitboard64 bits) {
  bits |= (bits << 1); bits |= (bits << 2); bits |= (bits << 4);
  return bits;
}

static inline bitboard64 bitboard_fill_SE(bitboard64 bits) {
  bits |= (bits >> 7); bits |= (bits >> 14); bits |= (bits >> 28);
  return bits;
}

static inline bitboard64 bitboard_fill_S(bitboard64 bits) {
  bits |= (bits >> 8); bits |= (bits >> 16); bits |= (bits >> 32);
  return bits;
}

static inline bitboard64 bitboard_fill_SW(bitboard64 bits) {
  bits |= (bits >> 9); bits |= (bits >> 18); bits |= (bits >> 36);
  return bits;
}

static inline bitboard64 bitboard_fill_W(bitboard64 bits) {
  bits |= (bits >> 1); bits |= (bits >> 2); bits |= (bits >> 4);
  return bits;
}

static inline bool bitboard_is_multi_bit(bitboard64 bits) {
  return (bits & (bits - 1));
}


#endif /* bitboard_h */
