//
//  Board.h
//  fornax2_core
//
//  Created by Anders on 26.06.2018.
//  Copyright © 2018 Avividi. All rights reserved.
//

#ifndef board_h
#define board_h

#include "definitions.h"

static inline bool  board_is_check(const Board* board) {
  return board->friendly->pieces[INHABITANT_KINGS] & board->hostile->attackMask;
}

static inline bool board_is_legal(const Board* board) {
  return board->hostile->pieces[INHABITANT_KINGS] & board->friendly->attackMask;
}

static inline bool board_is_repetition(const Board* board) {
  if (board->copyState->halfMoveCount >= 100) return true;
  for (int i = 0; i < REPETITION_HISTORY_SIZE; ++i) {
    uint32_t truncatedHash = (uint32_t) board->copyState->hash;
    if (board->copyState->history[i] == truncatedHash) return true;
  };
  return false;
}

inhabitant board_get_inhabitant(const Board* board, square square, bool isWhite);
void board_set_inhabitant(Board* board, bool isWhite, square square, inhabitant inhabitant);

inhabitant board_get_captured(const Board* board, square pos);
bitboard64 board_pieces_mask(bitboard64* pieceArray);

eval board_count_material(const BoardSide* side);
eval board_count_positional(const BoardSide* side, const piecetable* table);
eval board_count_phase(const BoardSide* side);

void board_set_friendlies_and_hostiles(Board* board);
void board_copy(const Board* original, Board* copy);
void board_flip(const Board* board, Board* copy);
void board_clear(Board* board);
Board* board_allocate(void);
void board_free(Board* board);

#define BOARD_COPY(name, original) { 0 };\
BoardCopiedState aux_##name; \
name.copyState = &aux_##name; \
board_copy(original, &name)

#endif /* board_h */
