//
//  definitions.h
//  fornax2_core
//
//  Created by Anders on 22/04/2019.
//  Copyright © 2019 Avividi. All rights reserved.
//

#ifndef definitions_h
#define definitions_h

#include <stdint.h>
#include <stdbool.h>
#include <limits.h>

#define STARTPOS_FEN "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1"

typedef uint64_t ttkey;
typedef uint8_t tttype;
typedef int16_t eval;
typedef int8_t piecetable[6][64];
typedef uint64_t bitboard64;
typedef int8_t square;
typedef uint32_t move;
/*
 0-5 origin
 6-11 dest
 12-14 originpiece
 15-17 destpiece
 18-20 capture
 21-28 value
 */

#define SEARCH_DEPTH_MAX (UINT8_MAX)
#define REPETITION_HISTORY_SIZE 16

typedef enum {
  INHABITANT_KINGS = 0,
  INHABITANT_PAWNS,
  INHABITANT_KNIGHTS,
  INHABITANT_BISHOPS,
  INHABITANT_ROOKS,
  INHABITANT_QUEENS
} inhabitant;

typedef struct {
  int x, y;
} Square;

typedef int8_t direction;
#define DIR_N 0
#define DIR_NE 1
#define DIR_E 2
#define DIR_SE 3
#define DIR_S 4
#define DIR_SW 5
#define DIR_W 6
#define DIR_NW 7

static const Square CARDINALS[8] = {
  {0, 1},
  {1, 1},
  {1, 0},
  {1, -1},
  {0, -1},
  {-1, -1},
  {-1, 0},
  {-1, 1}
};

typedef struct {
  square origin;
  inhabitant piece;
  bitboard64 moves;
} Movemap;

typedef struct {
  inhabitant piece;
  bitboard64 attacks;
} Attackmap;


typedef struct {
  ttkey hash;
  
  bool fcq;
  bool fck;
  bool hcq;
  bool hck;
  bitboard64 enPassant;
  
  uint8_t halfMoveCount;
  uint32_t history[REPETITION_HISTORY_SIZE];
  
} BoardCopiedState;

typedef struct {
  bitboard64 pieces[6];
  bitboard64 pieceMask;
  bitboard64 attackMask;
  Attackmap attackmaps[32];
  uint8_t attackmapCount;
  
  eval eval_material;
  eval eval_mg_positional;
  eval eval_eg_positional;
  uint8_t game_phase;
  
  const piecetable* mg_piecetable;
  const piecetable* eg_piecetable;
} BoardSide;


typedef struct {
  bool isWhiteMove;
  bool isCheck;
  
  BoardSide white;
  BoardSide black;
  BoardSide* friendly;
  BoardSide* hostile;
  
  Movemap movemaps[32];
  uint8_t movemapCount;
  
  BoardCopiedState* copyState;
  uint8_t ply;

  bool unchangedPawnStructure;
  eval pawnstucture_mg;
  eval pawnstucture_eg;
  
  inhabitant capturemap[64];
  
} Board;

typedef struct {
  move move;
  eval eval;
  long duration_ms;
} SearchResult;

#define SQUARE_NONE -1

#define SQUARE_A1 0
#define SQUARE_B1 1
#define SQUARE_C1 2
#define SQUARE_D1 3
#define SQUARE_E1 4
#define SQUARE_F1 5
#define SQUARE_G1 6
#define SQUARE_H1 7

#define SQUARE_A8 56
#define SQUARE_B8 57
#define SQUARE_C8 58
#define SQUARE_D8 59
#define SQUARE_E8 60
#define SQUARE_F8 61
#define SQUARE_G8 62
#define SQUARE_H8 63

#define TT_TYPE_EXACT 0 << 6
#define TT_TYPE_ALPHA 1 << 6
#define TT_TYPE_BETA 2 << 6

#define BB_CENTERMASK4 0x1818000000
#define BB_CENTERMASK16 0x3c3c3c3c0000
#define BB_RANK8 0xff00000000000000
#define BB_RANK7 0x00ff000000000000
#define BB_RANK6 0x0000ff0000000000
#define BB_RANK5 0x000000ff00000000
#define BB_RANK4 0x00000000ff000000
#define BB_RANK3 0x0000000000ff0000
#define BB_RANK2 0x000000000000ff00
#define BB_RANK1 0x00000000000000ff

#define BB_FILE_H 0x8080808080808080
#define BB_FILE_G 0x4040404040404040
#define BB_FILE_F 0x2020202020202020
#define BB_FILE_E 0x1010101010101010
#define BB_FILE_D 0x0808080808080808
#define BB_FILE_C 0x0404040404040404
#define BB_FILE_B 0x0202020202020202
#define BB_FILE_A 0x0101010101010101

static const bitboard64 BB_RANKS[8] = {
  BB_RANK1,
  BB_RANK2,
  BB_RANK3,
  BB_RANK4,
  BB_RANK5,
  BB_RANK6,
  BB_RANK7,
  BB_RANK8
};

#define BB_ODD_RANKS 0xAAAAAAAAAAAAAAAA
#define BB_EVEN_RANKS 0x5555555555555555

#define BB_CASTLING_DESTINATIONS 0x4400000000000044
#endif /* definitions_h */
