//
//  move.c
//  fornax2_core
//
//  Created by Anders on 26/06/2019.
//  Copyright © 2019 Avividi. All rights reserved.
//

#include "move.h"
#include "square.h"

move move_from_uci(const char* ucimove, const int length) {
  move m = 0;
  Square origin = {square_from_file(*ucimove), square_from_rank(*(ucimove + 1))};
  Square dest = {square_from_file(*(ucimove + 2)), square_from_rank(*(ucimove + 3))};
  m = move_set_origin(m, Square_flatten(origin));
  m = move_set_destination(m, Square_flatten(dest));
  m = move_set_origin_piece(m, 0);
  m = move_set_destination_piece(m, 0);
  m = move_set_capture(m, 0);
  
  if (length == 5) {
    int promotion = square_piece_from_symbol(*(ucimove + 4));
    if ( (int) promotion < 0) promotion = 0;
    m = move_set_destination_piece(m, promotion);
  }
  return m;
}

static int8_t move_calculate_promotion_value(inhabitant originpiece, inhabitant destinationpiece) {
  int8_t opv = MOVE_PIECEVALUE[originpiece];
  int8_t dpv = MOVE_PIECEVALUE[destinationpiece];
  int8_t promotionbonus = dpv - opv;
  
  
  return promotionbonus;
}


move move_annotate_order_value(move m, const Board* board) {
  inhabitant originpiece = MOVE_ORIGIN_PIECE(m);
  inhabitant destinationpiece = MOVE_DESTINATION_PIECE(m);
  inhabitant capturepiece = MOVE_CAPTURE(m);
  
  square origin = MOVE_ORIGIN(m);
  square destination = MOVE_DESTINATION(m);
  
  int8_t capture = MOVE_CAPTUREVALUE[capturepiece];
  int8_t promotion = move_calculate_promotion_value(originpiece, destinationpiece);
  
  int8_t destinationpenalty = originpiece == INHABITANT_PAWNS
  ? 0 : move_destination_penalty_value(originpiece, BITBOARD_TEST(board->hostile->attackMask, destination));
  int8_t evasionbonus = 0;
//  int8_t evasionbonus = originpiece == INHABITANT_PAWNS
//  ? 0 : move_evasion_value(originpiece, BITBOARD_TEST(board->hostile->attackMask, origin));
  
  int8_t protectionscore = evasionbonus + destinationpenalty;
  
  assert((int) capture + (int) promotion + (int) protectionscore < (int) INT8_MAX);
  assert((int) capture + (int) promotion + (int) protectionscore > (int)INT8_MIN);
  
  int8_t value = capture + promotion + protectionscore;
  
  const piecetable* table = board->isWhiteMove ? &WHITE_MOVE_PIECETABLE : &BLACK_MOVE_PIECETABLE;
  
  int8_t originvalue = (*table)[originpiece][origin];
  int8_t destinationvalue = (*table)[destinationpiece][destination];
  
  assert(value - originvalue < INT8_MAX);
  assert(value - originvalue > INT8_MIN);
  value -= originvalue;
  assert(value + destinationvalue < INT8_MAX);
  assert(value + destinationvalue > INT8_MIN);
  value += destinationvalue;
  
  move result = move_set_value(m, value);
  return result;
}

move move_full_from_uci(const Board* board, const char* ucimove, int length) {
  move m = move_from_uci(ucimove, length);
  inhabitant piece = board_get_inhabitant(board, MOVE_ORIGIN(m), board->isWhiteMove);
  inhabitant capture = board_get_captured(board, MOVE_DESTINATION(m));
  if (piece == INHABITANT_PAWNS && bitboard_test(board->copyState->enPassant, MOVE_DESTINATION(m))) {
    capture = INHABITANT_PAWNS;
  }
  m = move_set_capture(m, capture);
  m = move_set_origin_piece(m, piece);
  if (MOVE_DESTINATION_PIECE(m) == 0) {
    m = move_set_destination_piece(m, piece);
  }
  
  m = move_annotate_order_value(m, board);
  
  return m;
}

