//
//  move.h
//  fornax2_core
//
//  Created by Anders on 26/06/2019.
//  Copyright © 2019 Avividi. All rights reserved.
//

#ifndef move_h
#define move_h

#include "square.h"
#include "definitions.h"
#include "board.h"
#include "bitboard.h"
#include "move_piecetables.h"
#include "../eval/eval_definitions.h"

#define MOVE_WCQ 476053636
#define MOVE_WCK 476053892
#define MOVE_BCQ 476057276
#define MOVE_BCK 476057532


//#define MOVE_BCQ 0x200EBC
//#define MOVE_BCK 0x200FBC
//#define MOVE_WCQ 0x200084
//#define MOVE_WCK 0x200184

//#define MOVE_WCK 388
//#define MOVE_BCK 4028
//#define MOVE_WCQ 132
//#define MOVE_BCQ 3772
#define MOVE_SYMBOLIC_NULL 0x100000

#define MOVE_ORIGIN(move) (move & 0x3F)
#define MOVE_DESTINATION(move) ( move >> 6 & 0x3F)
#define MOVE_ORIGIN_PIECE(move) ( move >> 12 & 0x7)
#define MOVE_DESTINATION_PIECE(move) (move >> 15 & 0x7)
#define MOVE_CAPTURE(move) (move >> 18 & 0x7)
#define MOVE_VALUE(move) ((int8_t) (move >> 21 & 0xFF))
#define MOVE_IS_PROMOTION(move) (MOVE_ORIGIN_PIECE(move) != MOVE_DESTINATION_PIECE(move))

static inline move move_create(square origin,
                               square destination,
                               inhabitant originpiece,
                               inhabitant destpiece,
                               inhabitant capture,
                               int8_t value) {
  assert(origin >= 0 && origin < 64);
  assert(destination >= 0 && destination < 64);
  assert(originpiece >= 0 && originpiece <= 5);
  assert(destpiece >= 0 && destpiece <= 5);
  assert(capture >= 0 && capture <= 5);
  move m = 0
  ^ origin
  ^ (destination << 6)
  ^ (originpiece << 12)
  ^ (destpiece << 15)
  ^ (capture << 18)
  ^ (value << 21);
  return m;
}

static inline move move_set_origin(move m, square pos) {
  assert(pos >= 0 && pos < 64);
  return (m & 0xFFFFFFC0) ^ pos;
}
static inline move move_set_destination(move m, square pos) {
  assert(pos >= 0 && pos < 64);
  return (m & 0xFFFFF03F) ^ (pos << 6);
}
static inline move move_set_origin_piece(move m, inhabitant piece) {
  assert(piece >= 0 && piece <= 5);
  return (m & 0xFFFF8FFF) ^ (piece << 12);
}
static inline move move_set_destination_piece(move m, inhabitant piece) {
  assert(piece >= 0 && piece <= 5);
  return (m & 0xFFFC7FFF) ^ (piece << 15);
}
static inline move move_set_capture(move m, inhabitant piece) {
  assert(piece >= 0 && piece <= 5);
  return (m & 0xFFE3FFFF) ^ (piece << 18);
}
static inline move move_set_value(move m, int8_t value) {
  uint8_t bits = value;
  return (m & 0xE01FFFFF) ^ (bits << 21);
}
static inline move move_add_value(move m, int8_t value) {
  return move_set_value(m, MOVE_VALUE(m) + value);
}

static inline bool move_is_castling(move m) {
  int originx = SQUARE_X(MOVE_ORIGIN(m));
  int destinationx = SQUARE_X(MOVE_DESTINATION(m));
  return (MOVE_ORIGIN_PIECE(m) == INHABITANT_KINGS) & (originx == 4) & (destinationx == 2 || destinationx == 6);
}

static inline bool move_is_irreversible(move m) {
  return MOVE_CAPTURE(m) | (MOVE_ORIGIN_PIECE(m) == INHABITANT_PAWNS)
  | (m == MOVE_WCQ) | (m == MOVE_WCK) | (m == MOVE_BCQ) | (m == MOVE_BCK);
}

static inline int8_t move_destination_penalty_value(inhabitant originpiece, bool isProtectedSquare) {
  return - isProtectedSquare * MOVE_PIECEVALUE[originpiece] / 2;
}

static inline int8_t move_evasion_value(inhabitant originpiece, bool isProtectedSquare) {
  return isProtectedSquare * MOVE_PIECEVALUE[originpiece] / 4;
}


move move_annotate_order_value(move m, const Board* board);

move move_from_uci(const char* ucimove, const int length);
move move_full_from_uci(const Board* board, const char* ucimove, int length);


#endif /* move_h */
