//
//  eval_innner.inc
//  fornax2_core
//
//  Created by Anders on 24/03/2020.
//  Copyright © 2020 Avividi. All rights reserved.
//

#include "eval_definitions.h"
#include "eval_pawn_structure.inc.h"
#include "../board/bitboard.h"
#include "../boardprint.h"
#include "../board/board.h"
#include "../moving/movegen.h"
#include "../moving/oracle.h"
#include "../genericutils/ints.h"

#include "../boardprint.h"
#include <stdio.h>

#include "eval_debug_print.h"

static inline eval taper(uint8_t phase, eval mg, eval eg) {
  uint8_t max = PHASEVALUE_INITIAL;
  return (mg * phase + eg * (max - phase)) / max;
}

static inline eval eval_material(const BoardSide* white, const BoardSide* black) {
  eval w_mat = ED_ADD_WHITE("material", white->eval_material);
  eval b_mat = ED_ADD_BLACK("material", black->eval_material);
  eval mat = w_mat - b_mat;
  return mat;
}


static inline void eval_bishop_pair(eval* mg, eval* eg, const BoardSide* white, const BoardSide* black) {
  bool w_pair = bitboard_is_multi_bit(white->pieces[INHABITANT_BISHOPS]);
  eval w_mg_bp = ED_ADD_WHITE_MG("bishop pair", w_pair * 15);
  eval w_eg_bp = ED_ADD_WHITE_EG("bishop pair", w_pair * 35);
  
  bool b_pair = bitboard_is_multi_bit(black->pieces[INHABITANT_BISHOPS]);
  eval b_mg_bp = ED_ADD_BLACK_MG("bishop pair", b_pair * 15);
  eval b_eg_bp = ED_ADD_BLACK_EG("bishop pair", b_pair * 35);
  
  eval mg_diff = ED_ADD_DIFF_MG("bishop pair", w_mg_bp - b_mg_bp);
  eval eg_diff = ED_ADD_DIFF_MG("bishop pair", w_eg_bp - b_eg_bp);
  
  *mg += mg_diff;
  *eg += eg_diff;
}

static inline void eval_positional(eval* mg, eval* eg, const BoardSide* white, const BoardSide* black) {
  eval w_mg_positional = ED_ADD_WHITE_MG("positional", white->eval_mg_positional);
  eval w_eg_positional = ED_ADD_WHITE_EG("positional", white->eval_eg_positional);
  
  eval b_mg_positional = ED_ADD_BLACK_MG("positional", black->eval_mg_positional);
  eval b_eg_positional = ED_ADD_BLACK_EG("positional", black->eval_eg_positional);
  
  eval mg_positional = (w_mg_positional - b_mg_positional);
  mg_positional = ED_ADD_DIFF_MG("positional", mg_positional);
  eval eg_positional = (w_eg_positional - b_eg_positional);
  eg_positional = ED_ADD_DIFF_EG("positional", eg_positional);
  
  *mg += mg_positional;
  *eg += eg_positional;
}

static inline eval piece_activity(bitboard64 attacks, eval activityValue) {
  eval pieceActivity = BITBOARD_COUNT(attacks);
  pieceActivity += BITBOARD_COUNT(attacks & BB_CENTERMASK16);
  pieceActivity += BITBOARD_COUNT(attacks & BB_CENTERMASK4);
  pieceActivity *= activityValue;
  return pieceActivity;
}

static inline eval piece_connectivity(bitboard64 attacks, eval activityValue) {
  return BITBOARD_COUNT(attacks) * activityValue;
}

static inline eval piece_threats(bitboard64 attacks, eval activityValue, const bitboard64* hostiles) {
  eval threats = BITBOARD_COUNT(attacks & hostiles[INHABITANT_KNIGHTS]) * THREATVALUE_KNIGHTS;
  threats += BITBOARD_COUNT(attacks & hostiles[INHABITANT_ROOKS]) * THREATVALUE_ROOKS;
  threats += BITBOARD_COUNT(attacks & hostiles[INHABITANT_BISHOPS]) * THREATVALUE_BISHOPS;
  threats += BITBOARD_COUNT(attacks & hostiles[INHABITANT_QUEENS]) * THREATVALUE_QUEENS;
  threats += BITBOARD_COUNT(attacks & hostiles[INHABITANT_PAWNS]) * THREATVALUE_PAWNS;
  threats *= activityValue;
  return threats;
}

static inline eval king_threats(bitboard64 attacks, bitboard64 kingNeighborSquares) {
  return BITBOARD_COUNT(attacks & kingNeighborSquares);
}

static inline eval king_defenders(bitboard64 attacks, bitboard64 kingNeighborSquares) {
  return BITBOARD_COUNT(attacks & kingNeighborSquares);
}

static inline eval king_pawn_shield(bitboard64 kingNeighborSquares, const bitboard64* friendlies) {
  return BITBOARD_COUNT(kingNeighborSquares & friendlies[INHABITANT_PAWNS]);
}

static inline eval king_safety(eval defenders, eval attacks, eval pawnShield) {
  eval def = defenders - 5 + pawnShield * pawnShield;
  attacks = attacks * 2;
  eval diff = attacks - def;
  eval threat = diff > 0 ? (diff * diff) : diff / 2;
  return -threat;
}

static inline void eval_activity(eval* mg, eval* eg, const BoardSide* white, const BoardSide* black) {
  const bitboard64* whitepieces = white->pieces;
  const bitboard64* blackpieces = black->pieces;
  const bitboard64 w_king_neighbors = oracle_get_king_neighbor_vuln_squares(BITBOARD_FIND_TRAILING(*whitepieces));
  const bitboard64 b_king_neighbors = oracle_get_king_neighbor_vuln_squares(BITBOARD_FIND_TRAILING(*blackpieces));

  eval w_activity = 0;
  eval w_threats = 0;
  eval w_connectivity = 0;
  eval w_kingDefenders = 0;
  eval w_kingThreat = 0;
  eval w_kingPawnShield = king_pawn_shield(w_king_neighbors, whitepieces);

  eval b_activity = 0;
  eval b_threats = 0;
  eval b_connectivity = 0;
  eval b_kingDefenders = 0;
  eval b_kingThreat = 0;
  eval b_kingPawnShield = king_pawn_shield(b_king_neighbors, blackpieces);

  for (uint8_t i = 0; i < white->attackmapCount; ++i) {
    Attackmap set = white->attackmaps[i];
    eval activityValue = ACTIVITYVALUE[set.piece];
    bitboard64 attacks = set.attacks;
    bitboard64 hostileattacks = set.attacks & black->pieceMask;
    bitboard64 friendlyattaks = set.attacks & white->pieceMask;
    bitboard64 emptyattacks = attacks ^ (hostileattacks | friendlyattaks);
    
    w_activity += piece_activity(emptyattacks, activityValue);
    w_threats += piece_threats(hostileattacks, activityValue, blackpieces);
    w_connectivity += piece_connectivity(friendlyattaks, activityValue);
    w_kingDefenders += king_defenders(attacks, w_king_neighbors);
    b_kingThreat += king_threats(attacks, b_king_neighbors);
  }
  for (uint8_t i = 0; i < black->attackmapCount; ++i) {
    Attackmap set = black->attackmaps[i];
    eval activityValue = ACTIVITYVALUE[set.piece];
    bitboard64 attacks = set.attacks;
    bitboard64 hostileattacks = set.attacks & white->pieceMask;
    bitboard64 friendlyattaks = set.attacks & black->pieceMask;
    bitboard64 emptyattacks = attacks ^ (hostileattacks | friendlyattaks);
    
    b_activity += piece_activity(emptyattacks, activityValue);
    b_threats += piece_threats(hostileattacks, activityValue, whitepieces);
    b_connectivity += piece_connectivity(friendlyattaks, activityValue);
    b_kingDefenders += king_defenders(attacks, b_king_neighbors);
    w_kingThreat += king_threats(attacks, w_king_neighbors);
  }

  eval activity = ED_ADD_WHITE("activity", w_activity) - ED_ADD_BLACK("activity", b_activity);
  activity = ED_ADD_DIFF("activity", activity / 2);

  eval threats = ED_ADD_WHITE("threats", w_threats) - ED_ADD_BLACK("threats", b_threats);
  threats = ED_ADD_DIFF("threats", threats);
  
  eval connectivity = ED_ADD_WHITE("connectivity", w_connectivity) - ED_ADD_BLACK("connectivity", b_connectivity);
  connectivity = ED_ADD_DIFF("connectivity", connectivity / 4);

  eval w_kingSafety_mg = ED_ADD_WHITE_MG("king safety", king_safety(w_kingDefenders, w_kingThreat, w_kingPawnShield));
  eval b_kingSafety_mg = ED_ADD_BLACK_MG("king safety", king_safety(b_kingDefenders, b_kingThreat, b_kingPawnShield));
  eval kingSafety_mg = w_kingSafety_mg - b_kingSafety_mg;

  *mg += activity + threats + connectivity + kingSafety_mg;
  *eg += activity + threats + connectivity;

}

#ifndef EVAL_EVALUATE
#error "EVAL_EVALUATE not defined"
#endif

eval EVAL_EVALUATE(Board* board) {
  ED_START(board->isWhiteMove);
  const BoardSide* white = &board->white;
  const BoardSide* black = &board->black;
  
  eval material = eval_material(white, black);
  eval mg_total = material;
  eval eg_total = material;
  
  eval_positional(&mg_total, &eg_total, white, black);
  eval_activity(&mg_total, &eg_total, white, black);
  eval_pawn_structure(&mg_total, &eg_total, board);
  eval_bishop_pair(&mg_total, &eg_total, white, black);
  
  uint8_t phase = MAX_INT(white->game_phase, black->game_phase);
  eval total = taper(phase, mg_total, eg_total);
  ED_END(board, total);
  return board->isWhiteMove ? total : -total;
}

