//
//  eval_pawn_structure.h
//  fornax2_core
//
//  Created by Anders on 26/04/2020.
//  Copyright © 2020 Avividi. All rights reserved.
//

#ifndef eval_pawn_structure_h
#define eval_pawn_structure_h

#include "../board/bitboard.h"
#include "../board/definitions.h"
#include "../eval/eval_debug_print.h"

#include "../boardprint.h"

static bool testwhite[64] =
{
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 1, 0, 0, 0,
  0, 0, 0, 1, 0, 0, 0, 0,
  1, 0, 1, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
};

static bool testblack[64] =
{
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 1, 0, 0, 1, 0, 1,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 1,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
};

static inline bitboard64 fromarray(bool array[64]) {
  bitboard64 bits = 0;
  for (int i = 0; i < 64; ++i)
    if (array[i])
      BITBOARD_SET(bits, i);
  return BITBOARD_FLIP(bits);
}

static inline bitboard64 pawn_span_N(bitboard64 pawns) {
  return bitboard_fill_N(BITBOARD_SHIFT_N(pawns));
}

static inline bitboard64 pawn_span_S(bitboard64 pawns) {
  return bitboard_fill_S(BITBOARD_SHIFT_S(pawns));
}
 
static inline bitboard64 pawn_file_fill(bitboard64 pawns) {
  return bitboard_fill_S(pawns) | bitboard_fill_N(pawns);
}

static inline bitboard64 passed_pawns_white(bitboard64 pawns, bitboard64 opposingpawns) {
  bitboard64 opposing = pawn_span_S(opposingpawns);
  opposing |= BITBOARD_SHIFT_W(opposing) | BITBOARD_SHIFT_E(opposing);
  return pawns & ~opposing;
}

static inline bitboard64 passed_pawns_black(bitboard64 pawns, bitboard64 opposingpawns) {
  bitboard64 opposing = pawn_span_N(opposingpawns);
  opposing |= BITBOARD_SHIFT_W(opposing) | BITBOARD_SHIFT_E(opposing);
  return pawns & ~opposing;
}

static inline bitboard64 doubled_pawns(bitboard64 pawns) {
  return (pawn_span_N(pawns) | pawn_span_S(pawns)) & pawns;
}

static inline bitboard64 isolated_pawns(bitboard64 pawns) {
  bitboard64 filefill  = pawn_file_fill(pawns);
  return (pawns & ~BITBOARD_SHIFT_E(filefill)) & (pawns & ~BITBOARD_SHIFT_W(filefill));
}

static inline void p_test(void) {
  bitboard64 white = fromarray(testwhite);
  bitboard64 black = fromarray(testblack);
  boardprint_bitboard(passed_pawns_white(white, black));
  boardprint_bitboard(passed_pawns_black(black, white));
  boardprint_bitboard(doubled_pawns(white));
  boardprint_bitboard(doubled_pawns(black));
}

static inline void eval_pawn_structure(eval* mg, eval* eg, Board* board) {
  #ifndef EVAL_DEBUG_PRINT
  if (board->unchangedPawnStructure) {
    *mg += board->pawnstucture_mg;
    *eg += board->pawnstucture_eg;
    return;
  }
  #endif
  
  bitboard64 white = board->white.pieces[INHABITANT_PAWNS];
  bitboard64 black = board->black.pieces[INHABITANT_PAWNS];
  
  eval whitepassers = BITBOARD_COUNT(passed_pawns_white(white, black));
  eval blackpassers = BITBOARD_COUNT(passed_pawns_black(black, white));
  eval passerdiff = ED_ADD_WHITE("passers", whitepassers) - ED_ADD_BLACK("passers", blackpassers);
  *mg += ED_ADD_DIFF_MG("passers", passerdiff * 12);
  *eg += ED_ADD_DIFF_EG("passers", passerdiff * 24);
  
  eval whitedoubles = -BITBOARD_COUNT(doubled_pawns(white));
  eval blackdoubles = -BITBOARD_COUNT(doubled_pawns(black));
  eval doublediff = ED_ADD_WHITE("doubled", whitedoubles) - ED_ADD_BLACK("doubled", blackdoubles);
  *mg += ED_ADD_DIFF_MG("doubled", doublediff * 4);
  *eg += ED_ADD_DIFF_EG("doubled", doublediff * 8);
  
  eval whiteisolated = -BITBOARD_COUNT(isolated_pawns(white));
  eval blackisolated = -BITBOARD_COUNT(isolated_pawns(black));
  eval isolateddiff = ED_ADD_WHITE("isolated", whiteisolated) - ED_ADD_BLACK("isolated", blackisolated);
  board->pawnstucture_mg = ED_ADD_DIFF_MG("isolated", isolateddiff * 8);
  board->pawnstucture_eg = ED_ADD_DIFF_EG("isolated", isolateddiff * 16);
  
  *mg += board->pawnstucture_mg;
  *eg += board->pawnstucture_eg;
  board->unchangedPawnStructure = true;
}
#endif /* eval_pawn_structure_h */
