//
//  eval_test_suite.c
//  fornax2_core
//
//  Created by Anders on 24/03/2020.
//  Copyright © 2020 Avividi. All rights reserved.
//

#include <stdio.h>
#include <assert.h>

#include "eval_test_suite.h"
#include "eval.h"
#include "../fen/fen.h"
#include "../board/board.h"
#include "../boardprint.h"

static int fail;
static int success;
static int total;

static void test(const char* label, const char* fen, eval min, eval max) {
  BoardCopiedState aux;
  Board b;
  b.copyState = &aux;
  fen_to_board(&b, fen);
  
  printf("testing %s:\n", label);
  boardprint_fen(b);
  printf("\n");
  
  eval ev = eval_evaluate(&b);
  BoardCopiedState faux;
  Board c;
  c.copyState = &faux;
  board_flip(&b, &c);
  eval flipped = eval_evaluate(&c);
  total++;
         
  printf("eval =  %d (<%d, %d>)\n", ev, min, max);
  
  if (ev != flipped) {
    fail++;
    printf("BUG: flipped eval was different = %d\n", flipped);
  }
  else if (ev < min) {
    fail++;
    printf("FAIL: eval = %d below expected bounds = %d\n", ev, min);
  }
  else if (ev > max) {
    fail++;
    printf("FAIL: eval = %d above expected bounds = %d\n", ev, max);
  }
  else {
    success++;
  }
  
  printf("----------------------------------------------\n");
}

void eval_test_suite_standard(void) {
  fail = 0;
  success = 0;
  total = 0;
  test("startpos", "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1", 0, 0);
  test("pos1", "r1bqkb1r/ppp2ppp/2n1pn2/8/2PPp3/2N1B3/PP3PPP/R2QKBNR w KQkq - 2 6", -150, -10);
  test("pos2", "r7/ppk3p1/2pq1n1p/3p1B2/3P4/2N5/PPP2PPP/1R2R1K1 w - - 0 23", -150, -5);
 
  test("rd1-lagrave-caruana", "1rbqk2r/b1p2ppp/p1np1n2/Pp2p3/3PP3/1BP2N2/1P3PPP/RNBQ1RK1 w k - 1 11", 60, 200);
  test("rd1-ding-hao", "rnbqr1k1/ppb2ppp/2pp1n2/4p3/NPP5/P2P1NP1/4PPBP/R1BQ1RK1 w - - 1 11", 0, 50);
  test("rd1-giri-nepomniachtchi", "r1b1kb1r/1p1p1ppp/1q2p3/1p2n3/2P2Bn1/2N1P1PP/PP3P2/R2QKB1R w KQkq - 0 11", -550, 50);
  test("rd1-grischuk-alekseenko", "1rbq1rk1/ppp2pp1/2n4p/2b1p3/8/2PP1NP1/P1Q1PPBP/R1B2RK1 w - - 2 11", -50, 50);
  test("rd2-caruana-alekseenko", "rnbqr1k1/pp2bppp/5n2/2pP4/2P5/P1N5/1P2B1PP/R1BQK1NR w KQ - 3 11", -40, 40);
  test("rd2-nepomniachtchi-grischuk", "r1b1kb1r/ppp2pp1/2p5/4Pn1p/8/2N2N1P/PPP2PP1/R1B2RK1 w - - 0 11", 20, 100);
  test("rd2-hao-giri", "r1bqk2r/pp3pb1/2n2npp/1N1p4/3p3P/5NP1/PP2PPB1/R1BQ1RK1 w kq - 2 11", -50, 50);
  test("rd2-lagrave-ding", "r4rk1/1bpqbppp/p1np1n2/1p2p3/4P3/PB1P1N1P/1PP2PP1/RNBQR1K1 w - - 1 11", -50, 50);
  test("rd3-liren-caruana", "rn1q1rk1/pp3ppp/2p2n2/4N3/Pb1P4/2N2P2/1Pb1PKPP/R1BQ1B1R w - - 1 11", -30, 100);
  test("rd3-giri-lagrave", "rnb2rk1/pp2pp1p/3q2p1/2pP4/4P3/2P5/P2Q1PPP/R3KBNR w KQ - 3 11", 10, 100);
  test("rd3-grischuk-hao", "rn2k2r/ppp1q1pp/3bpn2/3p4/3P4/3B4/PPP1QPPP/RNB1K2R w KQkq - 0 11", -5, 100);
  test("rd3-alekseenko-nepomniachtchi", "rn2k2r/p1q1nppp/1p2p3/2ppP3/b2P3P/P1P2N2/2P1BPP1/R1BQK2R w KQkq - 4 11", -5, 60);
  test("FG1", "r2q2k1/5ppp/p2br3/1b5Q/3B1P2/2PR4/P1P3PP/5RK1 w - - - -", -400, -100);
  test("FG2", "3r2k1/R4Rp1/1p6/1P6/1NK1b3/7r/1P6/8 b - - - -", 0, 200);
  test("FG3", "1r4k1/2p1q1p1/2p2p1p/3p3P/p2PnB1R/4PQ2/P1P2PP1/2K5 w - - - -", -600, -100);
  test("FG4", "3r2k1/pp3pp1/2pb1q1p/P2p1b2/3P4/2N1B2r/1PPQBPR1/1R4K1 w - - 2 26", -200, 0);
  test("FG5", "1r4k1/4p1p1/4p2p/n7/N2pP2K/P5P1/1PR3P1/8 w - - 0 1", -100, 100);
  test("FG6", "r3kb1r/p2b1ppp/2p2n2/2p1N3/8/1PB1PP2/P1PP1P1P/R3K2R w KQkq - - -", -250, -100);
  test("FG7", "6k1/p3rp1p/2P5/3K4/8/5BP1/4P3/8 b - - 2 47", 50, 150);
  assert(fail + success == total);
  printf("EVAL TESTS %d/%d \n", success, total);
}
