//
//  main.c
//  fornax2_core
//
//  Created by Anders on 25.06.2018.
//  Copyright © 2018 Avividi. All rights reserved.
//

#include <stdio.h>
#include <stdlib.h>
#include "genericutils/iolog.h"
#include "genericutils/strings.h"
#include "moving/oracle.h"
#include "perft/perft_test_suite.h"
#include "tt/transpositions.h"
#include "supplementary/opening_book.h"
#include "uci.h"
#include "boardprint.h"
#include "board/move.h"

static void print_version() {
  printf("FORNAX chess engine by Anders Fuglseth (v");
  uci_printversion();
  printf(")\n");
  fflush(stdout);
}


int main(int argc, const char * argv[]) {
  if (argc > 1) {
    if (string_equals("-v", argv[1])) {
      print_version();
      return EXIT_SUCCESS;
    }
  }
  
  oracle_init();
  perftsuite_fast();
  transpositions_init();
  opening_book_init();
  uci_init();
  print_version();

  char start[128] = "position startpos";
  uci_handlecommand(start);

  char* input = malloc(sizeof(char) * 8192);

  bool next = 1;
  while (next) {
    iolog_fgets(input, 8192);
    next = uci_handlecommand(input);
  }

  free(input);
  uci_destroy();
  opening_book_destroy();
  transpositions_destroy();
  oracle_destroy();

  
  return EXIT_SUCCESS;
}

