//
//  makemove.c
//  fornax2_core
//
//  Created by Anders on 18/04/2019.
//  Copyright © 2019 Avividi. All rights reserved.
//

#include "makemove.h"
#include "movegen.h"
#include "../genericutils/require.h"
#include "../board/square.h"
#include "../board/bitboard.h"
#include "../board/square.h"
#include "../board/board.h"
#include "../board/move.h"
#include "../tt/zobrist.h"
#include "../boardprint.h"
#ifndef NDEBUG
#include <stdio.h>
#include "../boardprint.h"
#endif

static inline void makemove_set_piecetable(BoardSide* side, square square, inhabitant piece) {
  side->eval_mg_positional += (*side->mg_piecetable)[piece][square];
  side->eval_eg_positional += (*side->eg_piecetable)[piece][square];
}

static inline void makemove_clear_piecetable(BoardSide* side, square square, inhabitant piece) {
  side->eval_mg_positional -= (*side->mg_piecetable)[piece][square];
  side->eval_eg_positional -= (*side->eg_piecetable)[piece][square];
}

static inline void makemove_set(BoardSide* side, square square, inhabitant piece) {
  bitboard64* pieceboard = side->pieces + piece;
  assert(!BITBOARD_TEST(*pieceboard, square));
  BITBOARD_SET(*pieceboard, square);
  BITBOARD_SET(side->pieceMask, square);
  side->eval_material += PIECEVALUE[piece];
  side->game_phase += PHASEVALUE[piece];
  makemove_set_piecetable(side, square, piece);
}

static inline void makemove_clear(BoardSide* side, square square, inhabitant piece) {
  bitboard64* pieceboard = side->pieces + piece;
  assert(BITBOARD_TEST(*pieceboard, square));
  BITBOARD_CLEAR(*pieceboard, square);
  BITBOARD_CLEAR(side->pieceMask, square);
  side->eval_material -= PIECEVALUE[piece];
  side->game_phase -= PHASEVALUE[piece];
  makemove_clear_piecetable(side, square, piece);
}

static inline void setEnPassant (Board* board,
                                 square origin,
                                 square destination) {
  
  int start = 1 + board->isWhiteMove * 5;
  int end = 3 + board->isWhiteMove;
  bool set_passant = (SQUARE_Y(origin) == start) & (SQUARE_Y(destination) == end);
  square square = (origin & 7) + 16 + (board->isWhiteMove * 24);
  board->copyState->enPassant = (set_passant * (((bitboard64)1U) << square));
  board->copyState->hash ^= ZOBRIST_ENPASSANT(board->copyState->enPassant);
}

static inline void handleCastling (Board* board, BoardSide* side, move m) {
  bitboard64* rooks = side->pieces + INHABITANT_ROOKS;
  bitboard64* kings = side->pieces;
  BoardCopiedState* newstate = board->copyState;
  bitboard64* total = &side->pieceMask;
  
  switch (m) {
    case MOVE_WCQ:
      BITBOARD_CLEAR(*rooks, SQUARE_A1);
      BITBOARD_CLEAR(*total, SQUARE_A1);
      makemove_clear_piecetable(side, SQUARE_A1, INHABITANT_ROOKS);
      board->capturemap[SQUARE_A1] = 0;
      newstate->hash ^= ZOBRIST_PIECE(SQUARE_A1, INHABITANT_ROOKS, true);
      BITBOARD_SET(*rooks, SQUARE_D1);
      BITBOARD_SET(*total, SQUARE_D1);
      makemove_set_piecetable(side, SQUARE_D1, INHABITANT_ROOKS);
      board->capturemap[SQUARE_D1] = INHABITANT_ROOKS;
      newstate->hash ^= ZOBRIST_PIECE(SQUARE_D1, INHABITANT_ROOKS, true);
      BITBOARD_SET(*kings, SQUARE_E1);
      BITBOARD_SET(*kings, SQUARE_D1);
      return;
    case MOVE_WCK:
      BITBOARD_CLEAR(*rooks, SQUARE_H1);
      BITBOARD_CLEAR(*total, SQUARE_H1);
      makemove_clear_piecetable(side, SQUARE_H1, INHABITANT_ROOKS);
      board->capturemap[SQUARE_H1] = 0;
      newstate->hash ^= ZOBRIST_PIECE(SQUARE_H1, INHABITANT_ROOKS, true);
      BITBOARD_SET(*rooks, SQUARE_F1);
      BITBOARD_SET(*total, SQUARE_F1);
      makemove_set_piecetable(side, SQUARE_F1, INHABITANT_ROOKS);
      board->capturemap[SQUARE_F1] = INHABITANT_ROOKS;
      newstate->hash ^= ZOBRIST_PIECE(SQUARE_F1, INHABITANT_ROOKS, true);
      BITBOARD_SET(*kings, SQUARE_E1);
      BITBOARD_SET(*kings, SQUARE_F1);
      return;
    case MOVE_BCQ:
      BITBOARD_CLEAR(*rooks, SQUARE_A8);
      BITBOARD_CLEAR(*total, SQUARE_A8);
      makemove_clear_piecetable(side, SQUARE_A8, INHABITANT_ROOKS);
      board->capturemap[SQUARE_A8] = 0;
      newstate->hash ^= ZOBRIST_PIECE(SQUARE_A8, INHABITANT_ROOKS, false);
      BITBOARD_SET(*rooks, SQUARE_D8);
      BITBOARD_SET(*total, SQUARE_D8);
      makemove_set_piecetable(side, SQUARE_D8, INHABITANT_ROOKS);
      board->capturemap[SQUARE_D8] = INHABITANT_ROOKS;
      newstate->hash ^= ZOBRIST_PIECE(SQUARE_D8, INHABITANT_ROOKS, false);
      BITBOARD_SET(*kings, SQUARE_E8);
      BITBOARD_SET(*kings, SQUARE_D8);
      return;
    case MOVE_BCK:
      BITBOARD_CLEAR(*rooks, SQUARE_H8);
      BITBOARD_CLEAR(*total, SQUARE_H8);
      makemove_clear_piecetable(side, SQUARE_H8, INHABITANT_ROOKS);
      board->capturemap[SQUARE_H8] = 0;
      newstate->hash ^= ZOBRIST_PIECE(SQUARE_H8, INHABITANT_ROOKS, false);
      BITBOARD_SET(*rooks, SQUARE_F8);
      BITBOARD_SET(*total, SQUARE_F8);
      makemove_set_piecetable(side, SQUARE_F8, INHABITANT_ROOKS);
      board->capturemap[SQUARE_F8] = INHABITANT_ROOKS;
      newstate->hash ^= ZOBRIST_PIECE(SQUARE_F8, INHABITANT_ROOKS, false);
      BITBOARD_SET(*kings, SQUARE_E8);
      BITBOARD_SET(*kings, SQUARE_F8);
      return;
  }
}

static inline void unhandleCastling (Board* board, BoardSide* side, move m) {
  bitboard64* rooks = side->pieces + INHABITANT_ROOKS;
  bitboard64* total = &side->pieceMask;
  switch (m) {
    case MOVE_WCQ:
      BITBOARD_SET(*rooks, SQUARE_A1);
      BITBOARD_SET(*total, SQUARE_A1);
      makemove_set_piecetable(side, SQUARE_A1, INHABITANT_ROOKS);
      board->capturemap[SQUARE_A1] = INHABITANT_ROOKS;
      BITBOARD_CLEAR(*rooks, SQUARE_D1);
      BITBOARD_CLEAR(*total, SQUARE_D1);
      makemove_clear_piecetable(side, SQUARE_D1, INHABITANT_ROOKS);
      board->capturemap[SQUARE_D1] = 0;
      return;
    case MOVE_WCK:
      BITBOARD_SET(*rooks, SQUARE_H1);
      BITBOARD_SET(*total, SQUARE_H1);
      makemove_set_piecetable(side, SQUARE_H1, INHABITANT_ROOKS);
      board->capturemap[SQUARE_H1] = INHABITANT_ROOKS;
      BITBOARD_CLEAR(*rooks, SQUARE_F1);
      BITBOARD_CLEAR(*total, SQUARE_F1);
      makemove_clear_piecetable(side, SQUARE_F1, INHABITANT_ROOKS);
      board->capturemap[SQUARE_F1] = 0;
      return;
    case MOVE_BCQ:
      BITBOARD_SET(*rooks, SQUARE_A8);
      BITBOARD_SET(*total, SQUARE_A8);
      makemove_set_piecetable(side, SQUARE_A8, INHABITANT_ROOKS);
      board->capturemap[SQUARE_A8] = INHABITANT_ROOKS;
      BITBOARD_CLEAR(*rooks, SQUARE_D8);
      BITBOARD_CLEAR(*total, SQUARE_D8);
      makemove_clear_piecetable(side, SQUARE_D8, INHABITANT_ROOKS);
      board->capturemap[SQUARE_D8] = 0;
      return;
    case MOVE_BCK:
      BITBOARD_SET(*rooks, SQUARE_H8);
      BITBOARD_SET(*total, SQUARE_H8);
      makemove_set_piecetable(side, SQUARE_H8, INHABITANT_ROOKS);
      board->capturemap[SQUARE_H8] = INHABITANT_ROOKS;
      BITBOARD_CLEAR(*rooks, SQUARE_F8);
      BITBOARD_CLEAR(*total, SQUARE_F8);
      makemove_clear_piecetable(side, SQUARE_F8, INHABITANT_ROOKS);
      board->capturemap[SQUARE_F8] = 0;
      return;
  }
}

static inline void cleanupCastling (Board* board) {
  bitboard64* king = board->hostile->pieces;
  *king &= (BB_CASTLING_DESTINATIONS | ((bool)0 - (BITBOARD_COUNT(*king) == 1)));
}

static inline bitboard64 is_legal(bitboard64 attacks, bitboard64 kingNotToMove) {
  return !(attacks & kingNotToMove);
}

static inline bool makemove_calc_after_states(Board* board) {
  movegen_genmaps(board);
  
  bool legal = is_legal(board->friendly->attackMask, *board->hostile->pieces);
  cleanupCastling(board);
  if (!legal) return legal;
  movegen_genmaps_hostile_attacks(board);
  
#ifndef NDEBUG
  if (board->copyState->hash != zobrist_get(board)) {
    printf("\n");
    boardprint_board(*board);
    printf("\n");
    printf("%llu vs %llu\n", board->copyState->hash, zobrist_get(board));
    assert(false);
  }
#endif
  
  assert(board->friendly->pieceMask == board_pieces_mask(board->friendly->pieces));
  assert(board->friendly->eval_material ==  board_count_material(board->friendly));
  assert(board->friendly->eval_mg_positional == board_count_positional(board->friendly, board->friendly->mg_piecetable));
  assert(board->friendly->eval_eg_positional == board_count_positional(board->friendly, board->friendly->eg_piecetable));
  assert(board->friendly->game_phase == board_count_phase(board->friendly));
  
  assert(board->hostile->pieceMask == board_pieces_mask(board->hostile->pieces));
  assert(board->hostile->eval_material ==  board_count_material(board->hostile));
  assert(board->hostile->eval_mg_positional == board_count_positional(board->hostile, board->hostile->mg_piecetable));
  assert(board->hostile->eval_eg_positional == board_count_positional(board->hostile, board->hostile->eg_piecetable));
  assert(board->hostile->game_phase == board_count_phase(board->hostile));
  
  return legal;
}

static inline void switch_copy_state(Board* board, BoardCopiedState* newstate) {
  newstate->hash = board->copyState->hash;
  newstate->hash ^= ZOBRIST_ENPASSANT(board->copyState->enPassant);
  newstate->hash ^= ZOBRIST_WHITE_MOVE();
  newstate->fcq = board->copyState->hcq;
  newstate->fck = board->copyState->hck;
  newstate->hcq = board->copyState->fcq;
  newstate->hck = board->copyState->fck;
  newstate->enPassant = 0;
  board->copyState = newstate;
}

static inline void switch_sides(Board* board) {
  board->isWhiteMove = !board->isWhiteMove;
  board->friendly = board->isWhiteMove ? &board->white : &board->black;
  board->hostile = board->isWhiteMove ? &board->black : &board->white;
}

static inline void unfold_history(move m, const BoardCopiedState* oldstate, BoardCopiedState* newstate) {
  bool irreversible = move_is_irreversible(m);
  newstate->halfMoveCount = irreversible ? 0 : oldstate->halfMoveCount + 1;
  if (irreversible) {
    for (int i = 0; i < REPETITION_HISTORY_SIZE; ++i) {
      newstate->history[i] = 0;
    }
  }
  else {
    for (int i = 0; i < REPETITION_HISTORY_SIZE; ++i) {
      newstate->history[i] = oldstate->history[i];
    }
    uint32_t truncatedhash = (uint32_t) oldstate->hash;
    newstate->history[oldstate->halfMoveCount % REPETITION_HISTORY_SIZE] = truncatedhash;
  }
}


bool makemove_make(Board* board, BoardCopiedState* newstate, move m) {
  bool wasWhite = board->isWhiteMove;
  
  unfold_history(m, board->copyState, newstate);
  
  switch_copy_state(board, newstate);
  switch_sides(board);
  board->ply++;
  
  square origin = MOVE_ORIGIN(m);
  square destination = MOVE_DESTINATION(m);
  square origin_piece = MOVE_ORIGIN_PIECE(m);
  square destination_piece = MOVE_DESTINATION_PIECE(m);
  inhabitant capture = MOVE_CAPTURE(m);
  
  BoardSide* friendly = board->friendly;
  BoardSide* hostile = board->hostile;
  
  makemove_clear(hostile, origin, origin_piece);
  newstate->hash ^= ZOBRIST_PIECE(origin, origin_piece, wasWhite);
  board->capturemap[origin] = 0;
  
  makemove_set(hostile, destination, destination_piece);
  newstate->hash ^= ZOBRIST_PIECE(destination, destination_piece, wasWhite);
  
  switch (origin_piece) {
    case INHABITANT_KINGS://castling
      handleCastling(board, hostile, m);
      newstate->hash ^= newstate->hcq * ZOBRIST_CASTLE_Q(wasWhite);
      newstate->hcq = false;
      newstate->hash ^= newstate->hck * ZOBRIST_CASTLE_K(wasWhite);
      newstate->hck = false;
      break;
    case INHABITANT_PAWNS:
      setEnPassant(board, origin, destination);
      board->unchangedPawnStructure = false;
      break;
    case INHABITANT_ROOKS://castling
    {
      square square_rook_origin_fcq = wasWhite ? SQUARE_A1 : SQUARE_A8;
      bool can_lose_fcq = origin == square_rook_origin_fcq;
      newstate->hash ^= newstate->hcq * can_lose_fcq * ZOBRIST_CASTLE_Q(wasWhite);
      newstate->hcq = newstate->hcq & !can_lose_fcq;
      
      square square_rook_origin_fck = wasWhite ? SQUARE_H1 : SQUARE_H8;
      bool can_lose_fck = origin == square_rook_origin_fck;
      newstate->hash ^= newstate->hck * can_lose_fck * ZOBRIST_CASTLE_K(wasWhite);
      newstate->hck = newstate->hck & !can_lose_fck;
    }
  }
  
  if (capture) {
    bitboard64* captureboard = friendly->pieces + capture;
    square capturesquare = destination; //en passant logic
  
    switch (capture) {
      case INHABITANT_PAWNS: {
        board->unchangedPawnStructure = false;
        capturesquare += ((UINT8_MAX + BITBOARD_TEST(*captureboard, destination)) & ((board->isWhiteMove << 4) - 8)); //en passant logic
        board->capturemap[capturesquare] = 0;
        break;
      }
      case INHABITANT_ROOKS:
      {
        square square_rook_origin_hcq = wasWhite ? SQUARE_A8 : SQUARE_A1;
        bool can_lose_hcq = destination == square_rook_origin_hcq;//castling
        newstate->hash ^= newstate->fcq * can_lose_hcq * ZOBRIST_CASTLE_Q(board->isWhiteMove);
        newstate->fcq = newstate->fcq & !can_lose_hcq;
        
        square square_rook_origin_hck = wasWhite ? SQUARE_H8 : SQUARE_H1;
        bool can_lose_hck = destination == square_rook_origin_hck;//castling
        newstate->hash ^= newstate->fck * can_lose_hck * ZOBRIST_CASTLE_K(board->isWhiteMove);
        newstate->fck = newstate->fck & !can_lose_hck;
      }
      default: {}
    }
    makemove_clear(friendly, capturesquare, capture);
    board->copyState->hash ^= ZOBRIST_PIECE(capturesquare, capture, board->isWhiteMove);
  }
  
  board->capturemap[destination] = destination_piece;
  return makemove_calc_after_states(board);
}

void makemove_takeback(Board* board, BoardCopiedState* preservedState, move m) {
  
  board->copyState = preservedState;
  
  switch_sides(board);
  board->ply--;
  
  BoardSide* friendly = board->friendly;
  BoardSide* hostile = board->hostile;
  
  square origin = MOVE_ORIGIN(m);
  square destination = MOVE_DESTINATION(m);
  square origin_piece = MOVE_ORIGIN_PIECE(m);
  square destination_piece = MOVE_DESTINATION_PIECE(m);
  square capture = MOVE_CAPTURE(m);
  
  makemove_clear(friendly, destination, destination_piece);
  board->capturemap[destination] = 0;
  makemove_set(friendly, origin, origin_piece);
  board->capturemap[origin] = origin_piece;
  
  if (origin_piece == INHABITANT_KINGS) {
    unhandleCastling(board, friendly, m);
  }
  if (capture) {
    if (BITBOARD_TEST(board->copyState->enPassant, destination)) {
      destination = destination + (8 - board->isWhiteMove * 16);
    }
    assert(destination >= 0);
    assert(destination < 64);
    makemove_set(hostile, destination, capture);
    board->capturemap[destination] = capture;
  }
  
  assert(board->copyState->hash == zobrist_get(board));
}


bool makemove_nullmove_make(Board* board, BoardCopiedState* newstate) {
  unfold_history(MOVE_SYMBOLIC_NULL, board->copyState, newstate);
  switch_copy_state(board, newstate);
  switch_sides(board);
  board->ply++;
  return makemove_calc_after_states(board);
}

void makemove_nullmove_takeback(Board* board, BoardCopiedState* preservedState) {
  board->copyState = preservedState;
  switch_sides(board);
  board->ply--;
}
