//
//  movegen_pawns.inc.h
//  fornax2_core
//
//  Created by Anders on 01/05/2020.
//  Copyright © 2020 Avividi. All rights reserved.
//

#ifndef movegen_pawns_inc_h
#define movegen_pawns_inc_h

static inline move movegen_pawn_move(const piecetable* table, move base, square origin, square destination) {
  move m = move_set_origin(base, origin);
  int8_t moveValue = -(*table)[INHABITANT_PAWNS][origin];
  m = move_set_destination(m, destination);
  moveValue += (*table)[INHABITANT_PAWNS][destination];
  return move_add_value(m, moveValue);
}

static inline move movegen_pawn_capture(const Board* board, move m) {
  inhabitant capture = board->capturemap[MOVE_DESTINATION(m)];
  capture += (capture == 0) * INHABITANT_PAWNS;
  m = move_set_capture(m, capture);
  int8_t captureValue = MOVE_CAPTUREVALUE[capture];
  return move_add_value(m, captureValue);
}

static move movegen_create_promotion(move m, inhabitant piece, int8_t value) {
  m = move_add_value(m, value);
  m = move_set_destination_piece(m, piece);
  return m;
}

static inline move* movegen_pawn_promotions(move* pointer, move m) {
  *pointer++ = movegen_create_promotion(m, INHABITANT_QUEENS, MOVE_PROMOTIONVALUE_QUEENS);
  *pointer++ = movegen_create_promotion(m, INHABITANT_ROOKS, MOVE_PROMOTIONVALUE_ROOKS);
  *pointer++ = movegen_create_promotion(m, INHABITANT_KNIGHTS, MOVE_PROMOTIONVALUE_KNIGHTS);
  *pointer++ = movegen_create_promotion(m, INHABITANT_BISHOPS, MOVE_PROMOTIONVALUE_BISHOPS);
  return pointer;
}

static inline move* movegen_gen_white_pawn_pushes(const Board* board, move* pointer) {
  bitboard64 empty = ~(board->white.pieceMask | board->black.pieceMask);
  bitboard64 shifted = BITBOARD_SHIFT_S(empty);
  bitboard64 pawns = board->white.pieces[INHABITANT_PAWNS];
  bitboard64 push = pawns & shifted;
  
  move m = 0;
  m = move_set_origin_piece(m, INHABITANT_PAWNS);
  m = move_set_destination_piece(m, INHABITANT_PAWNS);
  m = move_set_value(m, MOVE_PIECEVALUE_NONE);
  
  square origin;
  while ((origin = bitboard_popTrailing(&push)) != SQUARE_NONE) {
    move next = movegen_pawn_move(&WHITE_MOVE_PIECETABLE, m, origin, origin + 8);
    if (SQUARE_Y(origin) == 6) pointer = movegen_pawn_promotions(pointer, next);
    else *pointer++ = next;
  }
  bitboard64 doublepush = pawns & shifted & BITBOARD_SHIFT_S(shifted) & BB_RANK2;
  while ((origin = bitboard_popTrailing(&doublepush)) != SQUARE_NONE) {
    *pointer++ = movegen_pawn_move(&WHITE_MOVE_PIECETABLE, m, origin, origin + 16);
  }
  
  return pointer;
}

static inline move* movegen_gen_black_pawn_pushes(const Board* board, move* pointer) {
  bitboard64 empty = ~(board->white.pieceMask | board->black.pieceMask);
  bitboard64 shifted = BITBOARD_SHIFT_N(empty);
  bitboard64 pawns = board->black.pieces[INHABITANT_PAWNS];
  bitboard64 push = pawns & shifted;
  
  move m = 0;
  m = move_set_origin_piece(m, INHABITANT_PAWNS);
  m = move_set_destination_piece(m, INHABITANT_PAWNS);
  m = move_set_value(m, MOVE_PIECEVALUE_NONE);
  
  square origin;
  while ((origin = bitboard_popTrailing(&push)) != SQUARE_NONE) {
    move next = movegen_pawn_move(&BLACK_MOVE_PIECETABLE, m, origin, origin - 8);
    if (SQUARE_Y(origin) == 1) pointer = movegen_pawn_promotions(pointer, next);
    else *pointer++ = next;
  }
  bitboard64 doublepush = pawns & shifted & BITBOARD_SHIFT_N(shifted) & BB_RANK7;
  while ((origin = bitboard_popTrailing(&doublepush)) != SQUARE_NONE) {
    *pointer++ = movegen_pawn_move(&BLACK_MOVE_PIECETABLE, m, origin, origin - 16);
  }
  
  return pointer;
}

static inline move* movegen_gen_white_pawn_captures(const Board* board, move* pointer) {
  bitboard64 hostile = board->black.pieceMask | board->copyState->enPassant;
  bitboard64 pawns = board->white.pieces[INHABITANT_PAWNS];
  bitboard64 up = BITBOARD_SHIFT_S(hostile);
  bitboard64 west = pawns & BITBOARD_SHIFT_E(up);
  move m = 0;
  m = move_set_origin_piece(m, INHABITANT_PAWNS);
  m = move_set_destination_piece(m, INHABITANT_PAWNS);
  
  square origin;
  while ((origin = bitboard_popTrailing(&west)) != SQUARE_NONE) {
    move next = movegen_pawn_move(&WHITE_MOVE_PIECETABLE,  m, origin, origin + 7);
    next = movegen_pawn_capture(board, next);
    if (SQUARE_Y(origin) == 6) pointer = movegen_pawn_promotions(pointer, next);
    else *pointer++ = next;
  }
  bitboard64 east = pawns & BITBOARD_SHIFT_W(up);
  while ((origin = bitboard_popTrailing(&east)) != SQUARE_NONE) {
    move next = movegen_pawn_move(&WHITE_MOVE_PIECETABLE,  m, origin, origin + 9);
    next = movegen_pawn_capture(board, next);
    if (SQUARE_Y(origin) == 6) pointer = movegen_pawn_promotions(pointer, next);
    else *pointer++ = next;
  }
  return pointer;
}

static inline move* movegen_gen_black_pawn_captures(Board* board, move* pointer) {
  bitboard64 hostile = board->white.pieceMask | board->copyState->enPassant;
  bitboard64 pawns = board->black.pieces[INHABITANT_PAWNS];
  bitboard64 up = BITBOARD_SHIFT_N(hostile);
  bitboard64 west = pawns & BITBOARD_SHIFT_E(up);
  move m = 0;
  m = move_set_origin_piece(m, INHABITANT_PAWNS);
  m = move_set_destination_piece(m, INHABITANT_PAWNS);
  
  square origin;
  while ((origin = bitboard_popTrailing(&west)) != SQUARE_NONE) {
    move next = movegen_pawn_move(&BLACK_MOVE_PIECETABLE,  m, origin, origin - 9);
    next = movegen_pawn_capture(board, next);
    if (SQUARE_Y(origin) == 1) pointer = movegen_pawn_promotions(pointer, next);
    else *pointer++ = next;
  }
  bitboard64 east = pawns & BITBOARD_SHIFT_W(up);
  while ((origin = bitboard_popTrailing(&east)) != SQUARE_NONE) {
    move next = movegen_pawn_move(&BLACK_MOVE_PIECETABLE,  m, origin, origin - 7);
    next = movegen_pawn_capture(board, next);
    if (SQUARE_Y(origin) == 1) pointer = movegen_pawn_promotions(pointer, next);
    else *pointer++ = next;
  }
  return pointer;
}

static inline move* movegen_gen_pawn_moves(Board* board, move* pointer) {
  if (board->isWhiteMove) {
    pointer = movegen_gen_white_pawn_pushes(board, pointer);
    pointer = movegen_gen_white_pawn_captures(board, pointer);
  }
  else {
    pointer =movegen_gen_black_pawn_pushes(board, pointer);
    pointer = movegen_gen_black_pawn_captures(board, pointer);
  }
  return pointer;
}
#endif /* movegen_pawns_inc_h */
