//
//  movepicker.c
//  fornax2_core
//
//  Created by Anders on 21/04/2020.
//  Copyright © 2020 Avividi. All rights reserved.
//

#include <stdlib.h>
#include <string.h>
#include "movepicker.h"
#include "../boardprint.h"

move killer1[UINT8_MAX] = { 0 };
move killer2[UINT8_MAX] = { 0 };

void movepick_set_killer(uint8_t ply, move m, move ttmove) {
  if (m == ttmove) return;
  if (MOVE_VALUE(m) >  MOVE_KILLER_STORE_THRESHOLD) return;
//  printf("### storing killer = ");
//  boardprint_debugmove(m);
  if (killer1[ply] != m) {
    killer2[ply] = killer1[ply];
    killer1[ply] = m;
  }
}

void movepick_clean_killers(void) {
  memset(killer1, 0, sizeof(killer1));
  memset(killer2, 0, sizeof(killer2));
}

static inline move verify_killer(move killer, move ttmove, Board* board) {
  return (killer != ttmove
          && BITBOARD_TEST(board->friendly->pieceMask, MOVE_ORIGIN(killer))
          && !BITBOARD_TEST(board->friendly->pieceMask, MOVE_DESTINATION(killer)))
  ? killer
  : 0;
}

static int8_t getvalue2(move m, move killer1, move killer2) {
  return MOVE_VALUE(m) + ((m == killer1) * MOVE_KILLER_VALUE) + ((m == killer2) * MOVE_KILLER_VALUE);
}

//static int8_t getvalue0(move m, move killer1, move killer2) {
//  return MOVE_VALUE(m);
//}


void movepick_start(Movepick* state, Board* board, move ttmove) {
  state->movecount = movegen_gen(board, state->moves);
  
  state->ttmove = ttmove;
  state->killer1 = verify_killer(killer1[board->ply], ttmove, board);
  state->killer2 = verify_killer(killer2[board->ply], ttmove, board);
  
//  state->moveValueFunction = killer1 || killer2 ? &getvalue2 : &getvalue0;
  
//  printf("ply = %d fen = ", board->ply);
//  boardprint_fen(*board);
//  printf("\ntt = ");
//  boardprint_debugmove(state->ttmove);
//  printf("killer1 = ");
//  boardprint_debugmove(state->killer1);
//  printf("killer2 = ");
//  boardprint_debugmove(state->killer2);
//  printf("---------------------\n");
//
//  {
//    Board copy = BOARD_COPY(copy, board);
//    Movepick copypick = *state;
//    move m;
//    while ((m = movepick_next(&copypick)) != 0) {
//      if (m == ttmove) printf("[tt] ");
//      if (m == killer1) printf("[k1] ");
//      if (m == killer2) printf("[k2] ");
//      boardprint_debugmove(m);
//    }
//  }
}

static inline move decrementmoves(Movepick* state, uint8_t bestindex) {
  move bestmove = state->moves[bestindex];
  state->moves[bestindex] = state->moves[state->movecount - 1];
  state->movecount--;
  return bestmove;
}

move movepick_next(Movepick* state) {
  if (state->ttmove) {
    for (uint8_t i = 0; i < state->movecount; ++i) {
      if (state->moves[i] != state->ttmove) continue;
      state->ttmove = 0;
      return decrementmoves(state, i);
    }
  }
  
  uint8_t bestindex = 0;
  int8_t bestscore = INT8_MIN;
  for (uint8_t i = 0; i < state->movecount; ++i) {
    move m = state->moves[i];
//    int8_t score = (*state->moveValueFunction)(m, state->killer1, state->killer2);
    int8_t score = getvalue2(m, state->killer1, state->killer2);
    if (score < bestscore) continue;
    bestscore = score;
    bestindex = i;
  }
  if (bestscore == INT8_MIN) return 0;
  
  return decrementmoves(state, bestindex);
}

void movepick_start_quiesce(Movepick* state, Board* board) {
  state->movecount = movegen_gen(board, state->moves);
}

move movepick_next_quiesce(Movepick* state) {
  uint8_t bestindex = 0;
  int8_t bestscore = 0;
  for (uint8_t i = 0; i < state->movecount; ++i) {
    int8_t score = MOVE_VALUE(state->moves[i]);
    if (score <= bestscore) continue;
    bestscore = score;
    bestindex = i;
  }
  if (bestscore <= 0) return 0;
  
  return decrementmoves(state, bestindex);
}


move movepick_root_next(RootMovePick* state) {
  
  if (state->ttmove != 0) {
    state->prevmove = state->ttmove;
    state->ttmove = 0;
    return state->prevmove;
  }
  
  move bestmove = 0;
  eval bestscore = INT16_MIN;
  for (uint8_t i = 0; i < state->movecount; ++i) {
    move candidate = state->moves[i];
    if (candidate == 0) continue;
    else if (candidate == state->prevmove) {
      state->moves[i] = 0;
    }
    else {
      RootMove rootcandidate = state->rootmoves[i];
      if (rootcandidate.score < bestscore) continue;
      bestscore = rootcandidate.score;
      bestmove = rootcandidate.move;
    }
  }
  state->prevmove = bestmove;
  return bestmove;
}

void movepick_root_list_reset(RootMovePick* state, move ttmove) {
  state->ttmove = ttmove;
  state->prevmove = 0;
  state->movecount = state->rootmovecount;
  for (int i = 0; i < state->movecount; ++i) {
    state->moves[i] = state->rootmoves[i].move;
  }
}

RootMovePick* movepick_root_list_init(void) {
  RootMovePick* root = malloc(sizeof(*root));
  
  for (int i = 0; i < 256; ++i) {
    RootMove empty = {0, EVAL_NONE};
    root->rootmoves[i] = empty;
  }
  root->rootmovecount = 0;
  root->movecount = 0;
  root->ttmove = 0;
  return root;
}

void movepick_root_list_add(RootMovePick* state, move m, eval score) {
  RootMove rm = {m, score};
  state->rootmoves[state->rootmovecount++] = rm;
}

void movepick_root_list_destroy(RootMovePick* state) {
  free(state);
}


