//
//  perft.c
//  fornax2_core
//
//  Created by Anders on 30.06.2018.
//  Copyright © 2018 Avividi. All rights reserved.
//

#include <stdio.h>
#include <pthread.h>
#include <unistd.h>
#include "perft.h"
#include "../genericutils/stopwatch.h"
#include "../moving/movegen.h"
#include "../moving/makemove.h"
#include "../board/definitions.h"
#include "../board/board.h"
#include "../board/bitboard.h"
#include "../boardprint.h"

long perft_simple(Board* board, uint8_t depth, bool printMoves) {
  if (depth == 0) {
    return 1;
  }
  
  move moves[256];
  uint8_t size = movegen_gen(board, moves);
  
  long count = 0;
  BoardCopiedState* prevstate = board->copyState;
  
  for (uint8_t i = 0; i < size; ++i) {
    move m = moves[i];

    BoardCopiedState nextstate;
    bool legal = makemove_make(board, &nextstate, m);
    if (legal) {
      long localCount = perft_simple(board, depth - 1, false);
      
      if (printMoves && localCount > 0) {
        boardprint_move(moves[i]);
        printf(" - %ld\n", localCount);
      }
      count += localCount;
    }
    makemove_takeback(board, prevstate, m);
  }
  
  return count;
}

typedef struct perft_args {
  long count;
  const Board* board;
  move* moves;
  long startMove;
  long endMove;
  uint8_t depth;
  int threadCount;
  bool printMoves;
} perft_args;

void *_perft_inthread(void *arg) {
  perft_args* args=(perft_args *)arg;
  
  Board board;
  BoardCopiedState cpaux;
  board.copyState = &cpaux;
  board_copy(args->board, &board);
  BoardCopiedState* prevstate = board.copyState;
  
  for (uint8_t i = args->startMove; i < args->endMove; ++i) {
    move m = args->moves[i];
    BoardCopiedState nextstate;
    bool legal = makemove_make(&board, &nextstate, m);
    if (legal) {
      long localCount = perft_simple(&board, args->depth - 1, false);

      if (args->printMoves && localCount > 0) {
           boardprint_move(args->moves[i]);
           printf(" - %ld\n", localCount);
         }
      args->count += localCount;
    }
    makemove_takeback(&board, prevstate, m);
  }
  return NULL;
}


long perft_threaded(Board* board, uint8_t depth, bool printMoves, int threadCount) {
  //  boardprint_board(*board);
  if (depth == 0) {
    return 1;
  }
  if (threadCount <= 1) {
    return perft_simple(board, depth, printMoves);
  }
  
  move moves[256];
  uint8_t moveListSize = movegen_gen(board, moves);
  
  if (moveListSize == 0) {
    return 0;
  }
  else if (moveListSize == 1) {
    BoardCopiedState newstate;
    Board copy = BOARD_COPY(copy, board);
    bool legal = makemove_make(&copy, &newstate, moves[0]);
    if (!legal) return 1;
    else return 1 + perft_threaded(board, depth - 1 , false, threadCount);
  }
  
  long endThread1 = moveListSize / 2;
  
  perft_args args1;
  args1.count = 0;
  args1.board = board;
  args1.moves = moves;
  args1.startMove = 0;
  args1.endMove = endThread1;
  args1.depth = depth;
  args1.threadCount = threadCount / 2;
  args1.printMoves = printMoves;
  pthread_t id1;
  pthread_create(&id1, NULL, _perft_inthread, (void*) &args1);
  
  perft_args args2;
  args2.count = 0;
  args2.board = board;
  args2.moves = moves;
  args2.startMove = endThread1;
  args2.endMove = moveListSize;
  args2.depth = depth;
  args2.threadCount = threadCount / 2;
  args2.printMoves = printMoves;
  _perft_inthread((void*) &args2);
  
  pthread_join(id1, NULL);
  
  return args1.count + args2.count;
}

void perft_run(const Board* board, uint8_t depth, int threadCount) {
  
  Board copy = BOARD_COPY(copy, board);
  
  stopwatch_wall_start();
  stopwatch_cpu_start();
  long count = perft_threaded(&copy, depth, true, threadCount);
  printf("%ld\n", count);
  stopwatch_wall_stop_print();
  stopwatch_cpu_stop_print();
}
