//
//  perft_test_suite.c
//  fornax2_core
//
//  Created by Anders on 27/04/2019.
//  Copyright © 2019 Avividi. All rights reserved.
//

#include <stdio.h>
#include <stdio.h>
#include <assert.h>

#include "perft.h"
#include "../genericutils/require.h"
#include "../genericutils/stopwatch.h"
#include "../fen/fen.h"
#include "../board/board.h"
#include "../boardprint.h"
#include "../moving/movegen.h"



static int fail;
static int success;
static int total;

static void test(const char* name, long should, Board* b, uint8_t depth) {
  
  Board cp;
  BoardCopiedState cpaux;
  cp.copyState = &cpaux;
  board_copy(b, &cp);
  
  long was = depth > 4
  ? perft_threaded(&cp, depth, false, 2)
  : perft_simple(&cp, depth, false);
  
  if (should != was) {
    printf("PERFT FAIL: '%s' depth=%d (should %lu, was %lu)\n", name, depth, should, was);
    fail++;
  }
  else {
    success++;
  }
  total++;
}

void perftsuite_case1(void) {
  printf("testing case1:\n");
  Board b = FEN_TO_BOARD(b, "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1");
  boardprint_board(b);
  test("1 startpos fen", 20, &b, 1);
  test("1 startpos fen", 400, &b, 2);
  test("1 startpos fen", 8902, &b, 3);
  test("1 startpos fen", 197281, &b, 4);
  test("1 startpos fen", 4865609, &b, 5);
}


void perftsuite_case2(void) {
  printf("testing case2:\n");
  Board b = FEN_TO_BOARD(b, "rnbqkbnr/pppp1ppp/8/4p3/4P3/8/PPPP1PPP/RNBQKBNR w KQkq e6 0 1");
  boardprint_board(b);
  test("2 e2e4 e7e5", 29, &b, 1);
  test("2 e2e4 e7e5", 835, &b, 2);
  test("2 e2e4 e7e5", 24825, &b, 3);
  test("2 e2e4 e7e5", 728887, &b, 4);
  test("2 e2e4 e7e5", 22273312, &b, 5);
}

void perftsuite_case3(void) {
  printf("testing case3:\n");
  Board b = FEN_TO_BOARD(b, "r3k2r/p1ppqpb1/bn2pnp1/3PN3/1p2P3/2N2Q1p/PPPBBPPP/R3K2R w KQkq - - -");
  boardprint_board(b);
  test("3 kiwi", 48, &b, 1);
  test("3 kiwi", 2039, &b, 2);
  test("3 kiwi", 97862, &b, 3);
  test("3 kiwi", 4085603, &b, 4);
  test("3 kiwi", 193690690, &b, 5);
}

void perftsuite_case4(void) {
  printf("testing case4:\n");
  Board b = FEN_TO_BOARD(b, "8/2p5/3p4/KP5r/1R3p1k/8/4P1P1/8 w - - - -");
  boardprint_board(b);
  test("4", 14, &b, 1);
  test("4", 191, &b, 2);
  test("4", 2812, &b, 3);
  test("4", 43238, &b, 4);
  test("4", 674624, &b, 5);
  test("4", 11030083, &b, 6);
  test("4", 178633661, &b, 7);
}

void perftsuite_case5(void) {
  printf("testing case5:\n");
  Board b = FEN_TO_BOARD(b, "r3k2r/Pppp1ppp/1b3nbN/nP6/BBP1P3/q4N2/Pp1P2PP/R2Q1RK1 w kq - 0 1");
  boardprint_board(b);
  test("5", 6, &b, 1);
  test("5", 264, &b, 2);
  test("5", 9467, &b, 3);
  test("5", 422333, &b, 4);
  test("5", 15833292, &b, 5);
  fen_to_board(&b, "r2q1rk1/pP1p2pp/Q4n2/bbp1p3/Np6/1B3NBn/pPPP1PPP/R3K2R b KQ - 0 1 ");
  boardprint_board(b);
  test("5", 6, &b, 1);
  test("5", 264, &b, 2);
  test("5", 9467, &b, 3);
  test("5", 422333, &b, 4);
  test("5", 15833292, &b, 5);
}

void perftsuite_case6(void) {
  printf("testing case6:\n");
  Board b = FEN_TO_BOARD(b, "rnbq1k1r/pp1Pbppp/2p5/8/2B5/8/PPP1NnPP/RNBQK2R w KQ - 1 8");
  boardprint_board(b);
  test("6 bugsy", 44, &b, 1);
  test("6 bugsy", 1486, &b, 2);
  test("6 bugsy", 62379, &b, 3);
  test("6 bugsy", 2103487, &b, 4);
  test("6 bugsy", 89941194, &b, 5);
}

void perftsuite_case7(void) {
  printf("testing case7:\n");
  Board b = FEN_TO_BOARD(b, "r4rk1/1pp1qppp/p1np1n2/2b1p1B1/2B1P1b1/P1NP1N2/1PP1QPPP/R4RK1 w - - 0 10 ");
  boardprint_board(b);
  test("7 altsy", 46, &b, 1);
  test("7 altsy", 2079, &b, 2);
  test("7 altsy", 89890, &b, 3);
  test("7 altsy", 3894594, &b, 4);
  test("7 altsy", 164075551, &b, 5);
}

void perftsuite_case8(void) {
  printf("testing case8:\n");
  Board b = FEN_TO_BOARD(b, "rnbqkbnr/pPppppp1/8/8/8/8/1PPPPPpP/RNBQKBNR w KQkq - 0 5");
  boardprint_board(b);
  test("8 promo", 31, &b, 1);
  test("8 promo", 957, &b, 2);
  test("8 promo", 28885, &b, 3);
  test("8 promo", 919300, &b, 4);
  test("8 promo", 29394458, &b, 5);
}

void perftsuite_case9(void) {
  printf("testing case9:\n");
  Board b = FEN_TO_BOARD(b, "8/2k1r3/3r4/8/8/1R6/2K3R1/8 w - -");
  boardprint_board(b);
  test("9 rooks", 29, &b, 1);
  test("9 rooks", 792, &b, 2);
  test("9 rooks", 21085, &b, 3);
  test("9 rooks", 558207, &b, 4);
  test("9 rooks", 14687292, &b, 5);
}

void perftsuite_case10(void) {
  printf("testing case10:\n");
  Board b = FEN_TO_BOARD(b, "8/2k5/b6b/8/8/1B6/2K5/6B1 w - -");
  boardprint_board(b);
  test("10 bishops", 18, &b, 1);
  test("10 bishops", 337, &b, 2);
  test("10 bishops", 6620, &b, 3);
  test("10 bishops", 123940, &b, 4);
  test("10 bishops", 2470048, &b, 5);
  test("10 bishops", 46688664, &b, 6);
}

void perftsuite_case11(void) {
  printf("testing case11:\n");
  Board b = FEN_TO_BOARD(b, "8/2k1nn2/8/8/8/8/2K2NN1/8 w - -");
  boardprint_board(b);
  test("11 knights", 18, &b, 1);
  test("11 knights", 359, &b, 2);
  test("11 knights", 6176, &b, 3);
  test("11 knights", 114053, &b, 4);
  test("11 knights", 1911342, &b, 5);
}

void perftsuite_case12(void) {
  printf("testing case12:\n");
  Board b = FEN_TO_BOARD(b, "8/2k2q2/3q4/8/8/8/1QK3Q1/8 w - -");
  boardprint_board(b);
  test("12 queens", 40, &b, 1);
  test("12 queens", 1478, &b, 2);
  test("12 queens", 46412, &b, 3);
  test("12 queens", 1572567, &b, 4);
}

void perftsuite_case13(void) {
  printf("testing case13:\n");
  Board b = FEN_TO_BOARD(b, "8/p1k4p/8/8/8/8/2K1P1P1/8 w - -");
  boardprint_board(b);
  test("13 pawns", 12, &b, 1);
  test("13 pawns", 144, &b, 2);
  test("13 pawns", 1535, &b, 3);
  test("13 pawns", 16290, &b, 4);
  test("13 pawns", 162078, &b, 5);
  test("13 pawns", 1611733, &b, 6);
  test("13 pawns", 15412327, &b, 7);
}

void perftsuite_case14(void) {
  printf("testing case14:\n");
  Board b = FEN_TO_BOARD(b, "5nkb/5p1p/Q2B2p1/8/P7/5BKP/6P1/6q1 w - - 4 33");
  boardprint_board(b);
  test("14 xray", 35, &b, 1);
  test("14 xray", 1024, &b, 2);
  test("14 xray", 31389, &b, 3);
  test("14 xray", 934526, &b, 4);
  test("14 xray", 28882969, &b, 5);
}


void perftsuite_all(void) {
  fail = 0;
  success = 0;
  total = 0;
  stopwatch_wall_start();
  
  perftsuite_case1();
  perftsuite_case2();
  perftsuite_case3();
  perftsuite_case4();
  perftsuite_case5();
  perftsuite_case6();
  perftsuite_case7();
  perftsuite_case8();
  perftsuite_case9();
  perftsuite_case10();
  perftsuite_case11();
  perftsuite_case12();
  perftsuite_case13();
  perftsuite_case14();
  
  stopwatch_wall_stop_print();
  
  assert(fail + success == total);
  require(success == total, "TESTS FAILED: %d/%d success \n", success, total);
  printf("TESTS OK %d/%d success \n", success, total);
}

void perftsuite_fast1(void) {
  Board b = FEN_TO_BOARD(b, "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1");
  test("1 startpos fen", 20, &b, 1);
  test("1 startpos fen", 400, &b, 2);
  test("1 startpos fen", 8902, &b, 3);
}

void perftsuite_fast2(void) {
  Board b = FEN_TO_BOARD(b, "rnbqkbnr/pppp1ppp/8/4p3/4P3/8/PPPP1PPP/RNBQKBNR w KQkq e6 0 1");
  test("2 e2e4 e7e5", 29, &b, 1);
  test("2 e2e4 e7e5", 835, &b, 2);
  test("2 e2e4 e7e5", 24825, &b, 3);
}

void perftsuite_fast3(void) {
  Board b = FEN_TO_BOARD(b, "r3k2r/p1ppqpb1/bn2pnp1/3PN3/1p2P3/2N2Q1p/PPPBBPPP/R3K2R w KQkq - - -");
  test("3 kiwi", 48, &b, 1);
  test("3 kiwi", 2039, &b, 2);
  test("3 kiwi", 97862, &b, 3);
}

void perftsuite_fast4(void) {
  Board b = FEN_TO_BOARD(b, "8/2p5/3p4/KP5r/1R3p1k/8/4P1P1/8 w - - - -");
  test("4", 14, &b, 1);
  test("4", 191, &b, 2);
  test("4", 2812, &b, 3);
  test("4", 43238, &b, 4);
}

void perftsuite_fast5(void) {
  Board b = FEN_TO_BOARD(b, "r3k2r/Pppp1ppp/1b3nbN/nP6/BBP1P3/q4N2/Pp1P2PP/R2Q1RK1 w kq - 0 1");
  test("5", 6, &b, 1);
  test("5", 264, &b, 2);
  test("5", 9467, &b, 3);
  fen_to_board(&b, "r2q1rk1/pP1p2pp/Q4n2/bbp1p3/Np6/1B3NBn/pPPP1PPP/R3K2R b KQ - 0 1 ");
  test("5", 6, &b, 1);
  test("5", 264, &b, 2);
  test("5", 9467, &b, 3);
}

void perftsuite_fast6(void) {
  Board b = FEN_TO_BOARD(b, "rnbq1k1r/pp1Pbppp/2p5/8/2B5/8/PPP1NnPP/RNBQK2R w KQ - 1 8");
  test("6 bugsy", 44, &b, 1);
  test("6 bugsy", 1486, &b, 2);
  test("6 bugsy", 62379, &b, 3);
}

void perftsuite_fast7(void) {
  Board b = FEN_TO_BOARD(b, "r4rk1/1pp1qppp/p1np1n2/2b1p1B1/2B1P1b1/P1NP1N2/1PP1QPPP/R4RK1 w - - 0 10 ");
  test("7 altsy", 46, &b, 1);
  test("7 altsy", 2079, &b, 2);
  test("7 altsy", 89890, &b, 3);
}

void perftsuite_fast8(void) {
  Board b = FEN_TO_BOARD(b, "rnbqkbnr/pPppppp1/8/8/8/8/1PPPPPpP/RNBQKBNR w KQkq - 0 5");
  test("7 altsy", 31, &b, 1);
  test("7 altsy", 957, &b, 2);
  test("7 altsy", 28885, &b, 3);
}

void perftsuite_fast9(void) {
  Board b = FEN_TO_BOARD(b, "8/2k1r3/3r4/8/8/1R6/2K3R1/8 w - -");
  test("9 rooks", 29, &b, 1);
  test("9 rooks", 792, &b, 2);
  test("9 rooks", 21085, &b, 3);
}

void perftsuite_fast10(void) {
  Board b = FEN_TO_BOARD(b, "8/2k5/b6b/8/8/1B6/2K5/6B1 w - -");
  test("10 bishops", 18, &b, 1);
  test("10 bishops", 337, &b, 2);
  test("10 bishops", 6620, &b, 3);
}

void perftsuite_fast11(void) {
  Board b = FEN_TO_BOARD(b, "8/2k1nn2/8/8/8/8/2K2NN1/8 w - -");
  test("11 knights", 18, &b, 1);
  test("11 knights", 359, &b, 2);
  test("11 knights", 6176, &b, 3);
}

void perftsuite_fast12(void) {
  Board b = FEN_TO_BOARD(b, "8/2k2q2/3q4/8/8/8/1QK3Q1/8 w - -");
  test("12 queens", 40, &b, 1);
  test("12 queens", 1478, &b, 2);
}

void perftsuite_fast13(void) {
  Board b = FEN_TO_BOARD(b, "8/p1k4p/8/8/8/8/2K1P1P1/8 w - -");
  test("13 pawns", 12, &b, 1);
  test("13 pawns", 144, &b, 2);
  test("13 pawns", 1535, &b, 3);
  test("13 pawns", 16290, &b, 4);
}

void perftsuite_fast14(void) {
  Board b = FEN_TO_BOARD(b, "5nkb/5p1p/Q2B2p1/8/P7/5BKP/6P1/6q1 w - - 4 33");
  test("14 xray", 35, &b, 1);
  test("14 xray", 1024, &b, 2);
  test("14 xray", 31389, &b, 3);
}



void perftsuite_fast(void) {
  fail = 0;
  success = 0;
  total = 0;
  
  perftsuite_fast1();
  perftsuite_fast2();
  perftsuite_fast3();
  perftsuite_fast4();
  perftsuite_fast5();
  perftsuite_fast6();
  perftsuite_fast7();
  perftsuite_fast8();
  perftsuite_fast8();
  perftsuite_fast9();
  perftsuite_fast10();
  perftsuite_fast11();
  perftsuite_fast12();
  perftsuite_fast13();
  perftsuite_fast14();
  
  assert(fail + success == total);
  require(success == total, "TESTS FAILED: %d/%d success \n", success, total);
}
