//
//  search_test_suite.c
//  fornax2_core
//
//  Created by Anders on 30/06/2019.
//  Copyright © 2019 Avividi. All rights reserved.
//

#include <stdio.h>
#include <string.h>
#include "search_test_suite.h"
#include "search.h"
#include "../tt/transpositions.h"
#include "../board/board.h"
#include "../board/definitions.h"
#include "../board/move.h"
#include "../boardprint.h"
#include "../fen/fen.h"
#include "../genericutils/stopwatch.h"

static void test(const char* name, char* should, const char* fen, uint8_t depth);

static void suite_1_fast() {
  
  test("QGA-2", "d1f3", "rnbqkbnr/p3pppp/8/1p6/2pP4/4P3/1P3PPP/RNBQKBNR w KQkq - 0 6", 5);
  test("WAC.001", "g3g6", "2rr3k/pp3pp1/1nnqbN1p/3pN3/2pP4/2P3Q1/PPB4P/R4RK1 w - - 0 1", 4);
  test("HAKMEM70", "g7g8n", "5B2/6P1/1p6/8/1N6/kP6/2K5/8 w - -", 7);
  test("CAST-1", "e8c8", "r3k1K1/2pq4/3p4/p2Pp1b1/1pP1Pp2/1B3PP1/PP1N4/RNBQ1R2 b q - 0 8", 5);
  test("DEF-2", "c6f3", "2q2rk1/2p3pp/N1b5/p7/1r3nP1/1PQR1P2/1BP5/3K2R1 b - -", 5);
  test("Saavedra", "c6c7", "8/8/1KP5/3r4/8/8/8/k7 w - - 0 1", 6);
  
}

static void suite_2_solids() {
  test("QGA-1", "a4b5", "rnbqkbnr/p3pppp/2p5/1p6/P1pP4/4P3/1P3PPP/RNBQKBNR w KQkq - 0 5", 9);
  test("QGA-2", "d1f3", "rnbqkbnr/p3pppp/8/1p6/2pP4/4P3/1P3PPP/RNBQKBNR w KQkq - 0 6", 9);
  test("WAC.001", "g3g6", "2rr3k/pp3pp1/1nnqbN1p/3pN3/2pP4/2P3Q1/PPB4P/R4RK1 w - - 0 1", 9);
  test("WAC.003", "e3g3", "5rk1/1ppb3p/p1pb4/6q1/3P1p1r/2P1R2P/PP1BQ1P1/5RKN w - - 0 1", 9);
  test("WAC.004", "h6h7", " r1bq2rk/pp3pbp/2p1p1pQ/7P/3P4/2PB1N2/PP3PPR/2KR4 w - - ", 9);
  test("UP-1", "g2h1n", "3rQB1k/p5b1/2b4p/8/8/1P6/P4KpP/2R2B1R b - - 0 30", 9);
  test("UP-2", "b2b1n", "8/8/8/8/8/2K5/1p4R1/2k5 b - - 3 77", 9);
  test("EP-1", "e4f3", "r3k2r/pbppqpb1/1pn3p1/7p/1N2pPn1/1PP4N/PB1P2PP/2QRKR2 b kq f3 0 1", 9);
  test("EP-2", "e5f6", "r2q1r2/pbpn2p1/1pn1p1k1/4PpN1/3P4/2PQ4/P4PPP/R1B1R1K1 w - f6 0 15", 9);
  test("CAST-1", "e8c8", "r3k1K1/2pq4/3p4/p2Pp1b1/1pP1Pp2/1B3PP1/PP1N4/RNBQ1R2 b q - 0 8", 9);
  test("CAST-2", "e1c1", "r7/p4ppp/B2kpn2/8/8/2P2P2/Pr3P1P/R3K2R w KQ - 0 18", 9);
  test("DEF-2", "c6f3", "2q2rk1/2p3pp/N1b5/p7/1r3nP1/1PQR1P2/1BP5/3K2R1 b - -", 9);
  test("STALE-1", "e2b2", "7k/R7/R3p1Q1/4p3/p3B3/P7/KP2r2q/2B5 b - - 0 1", 10);
  test("MATE#5", "d1e1", "4r3/2R2P1p/1p4pk/p7/5P2/8/1Q3KPP/3q4 b - - 1 35", 10);
  test("FINE#70", "a1b1", "8/k7/3p4/p2P1p2/P2P1P2/8/8/K7 w - -", 27);
  test("Sackmann", "e5f5", "8/8/2p/k1p1K/p1P/P/8/8 w - -", 27);
  test("Reti", "h8g7", "7K/8/k1P5/7p/8/8/8/8 w - - 0 1", 17);
  test("HAKMEM70", "g7g8n", "5B2/6P1/1p6/8/1N6/kP6/2K5/8 w - -", 9);
  test("Saavedra", "c6c7", "8/8/1KP5/3r4/8/8/8/k7 w - - 0 1", 11);
  test("BK.01", "d6d1", "1k1r4/pp1b1R2/3q2pp/4p3/2B5/4Q3/PPP2B2/2K5 b - -", 9);
  test("BK.12", "d7f5", "r3r1k1/ppqb1ppp/8/4p1NQ/8/2P5/PP3PPP/R3R1K1 b - -", 9);
  test("BK.14", "d1d2", "rnb2r1k/pp2p2p/2pp2p1/q2P1p2/8/1Pb2NP1/PB2PPBP/R2Q1RK1 w - -", 9);
  test("BK.15", "g4g7", "2r3k1/1p2q1pp/2b1pr2/p1pp4/6Q1/1P1PP1R1/P1PN2PP/5RK1 w - -", 9);
  test("BK.16", "d2e4", "r1bqkb1r/4npp1/p1p4p/1p1pP1B1/8/1B6/PPPN1PPP/R2Q1RK1 w kq -", 9);
  test("BK.21", "f5h6", "3rn2k/ppb2rpp/2ppqp2/5N2/2P1P3/1P5Q/PB3PPP/3RR1K1 w - -", 9);
}


static void suite_3_bkfull() {
  test("BK.01", "d6d1", "1k1r4/pp1b1R2/3q2pp/4p3/2B5/4Q3/PPP2B2/2K5 b - -", 9);
  test("BK.02", "d4d5", "3r1k2/4npp1/1ppr3p/p6P/P2PPPP1/1NR5/5K2/2R5 w - -", 9);//CANNOT
  test("BK.03", "f6f5", "2q1rr1k/3bbnnp/p2p1pp1/2pPp3/PpP1P1P1/1P2BNNP/2BQ1PRK/7R b - -", 9);//USTABIL
  test("BK.04", "e5e6", "rnbqkb1r/p3pppp/1p6/2ppP3/3N4/2P5/PPP1QPPP/R1B1KB1R w KQkq -", 9);
  test("BK.05", "c3d5", "r1b2rk1/2q1b1pp/p2ppn2/1p6/3QP3/1BN1B3/PPP3PP/R4RK1 w - -", 9);//CANNOT
  test("BK.06", "g5g6", "2r3k1/pppR1pp1/4p3/4P1P1/5P2/1P4K1/P1P5/8 w - -", 11);//CANNOT
  test("BK.07", "h5f6", "1nk1r1r1/pp2n1pp/4p3/q2pPp1N/b1pP1P2/B1P2R2/2P1B1PP/R2Q2K1 w - -", 9);//USTABIL
  test("BK.08", "f4f5", "4b3/p3kp2/6p1/3pP2p/2pP1P2/4K1P1/P3N2P/8 w - -", 11);
  test("BK.09", "f4f5", "2kr1bnr/pbpq4/2n1pp2/3p3p/3P1P1B/2N2N1Q/PPP3PP/2KR1B1R w - -", 9);//USTABIL
  test("BK.10", "c6e5", "3rr1k1/pp3pp1/1qn2np1/8/3p4/PP1R1P2/2P1NQPP/R1B3K1 b - -", 9);//CANNOT
  test("BK.11", "f2f4", "2r1nrk1/p2q1ppp/bp1p4/n1pPp3/P1P1P3/2PBB1N1/4QPPP/R4RK1 w - -", 9);//USTABIL
  test("BK.12", "d7f5", "r3r1k1/ppqb1ppp/8/4p1NQ/8/2P5/PP3PPP/R3R1K1 b - -", 9);
  test("BK.13", "b2b4", "r2q1rk1/4bppp/p2p4/2pP4/3pP3/3Q4/PP1B1PPP/R3R1K1 w - -", 9); //USTABIL
  test("BK.14", "d1d2", "rnb2r1k/pp2p2p/2pp2p1/q2P1p2/8/1Pb2NP1/PB2PPBP/R2Q1RK1 w - -", 9);
  test("BK.15", "g4g7", "2r3k1/1p2q1pp/2b1pr2/p1pp4/6Q1/1P1PP1R1/P1PN2PP/5RK1 w - -", 9);
  test("BK.16", "d2e4", "r1bqkb1r/4npp1/p1p4p/1p1pP1B1/8/1B6/PPPN1PPP/R2Q1RK1 w kq -", 9);
  test("BK.17", "h7h5", "r2q1rk1/1ppnbppp/p2p1nb1/3Pp3/2P1P1P1/2N2N1P/PPB1QP2/R1B2RK1 b - -", 9);//USTABIL
  test("BK.18", "c5b3", "r1bq1rk1/pp2ppbp/2np2p1/2n5/P3PP2/N1P2N2/1PB3PP/R1B1QRK1 b - -", 9);//USTABIL
  test("BK.19", "e8e4", "3rr3/2pq2pk/p2p1pnp/8/2QBPP2/1P6/P5PP/4RRK1 b - -", 9);
  test("BK.20", "g3g4", "r4k2/pb2bp1r/1p1qp2p/3pNp2/3P1P2/2N3P1/PPP1Q2P/2KRR3 w - -", 9);//USTABIL
  test("BK.21", "f5h6", "3rn2k/ppb2rpp/2ppqp2/5N2/2P1P3/1P5Q/PB3PPP/3RR1K1 w - -", 9);
  test("BK.22", "b7e4", "2r2rk1/1bqnbpp1/1p1ppn1p/pP6/N1P1P3/P2B1N1P/1B2QPP1/R2R2K1 b - -", 9);//USTABIL
  test("BK.23", "f7f6", "r1bqk2r/pp2bppp/2p5/3pP3/P2Q1P2/2N1B3/1PP3PP/R4RK1 b kq -", 9);//USTABIL
  test("BK.24", "f2f4", "r2qnrnk/p2b2b1/1p1p2pp/2pPpp2/1PP1P3/PRNBB3/3QNPPP/5RK1 w - -", 9);//CANNOT
  
  //  test("Behting", "d5c6", "8/8/7p/3KNN1k/2p4p/8/3P2p1/8 w - - ", 25);//CANNOT
}


static void suite_4_win_at_chess() {
  test("WAC.001", "g3g6", "2rr3k/pp3pp1/1nnqbN1p/3pN3/2pP4/2P3Q1/PPB4P/R4RK1 w - - 0 1", 7);
  test("WAC.002", "b3b2", "8/7p/5k2/5p2/p1p2P2/Pr1pPK2/1P1R3P/8 b - -", 13);//cannot (hard)
  test("WAC.003", "e3g3", "5rk1/1ppb3p/p1pb4/6q1/3P1p1r/2P1R2P/PP1BQ1P1/5RKN w - - 0 1", 7);
  test("WAC.004", "h6h7", "r1bq2rk/pp3pbp/2p1p1pQ/7P/3P4/2PB1N2/PP3PPR/2KR4 w - - ", 7);
  test("WAC.005", "c6c4", "5k2/6pp/p1qN4/1p1p4/3P4/2PKP2Q/PP3r2/3R4 b - -", 7);
  test("WAC.006", "b6b7", "7k/p7/1R5K/6r1/6p1/6P1/8/8 w - -", 7);
  test("WAC.007", "g4e3", "rnbqkb1r/pppp1ppp/8/4P3/6n1/7P/PPPNPPP1/R1BQKBNR b KQkq -", 7);
  test("WAC.008", "e7f7", "r4q1k/p2bR1rp/2p2Q1N/5p2/5p2/2P5/PP3PPP/R5K1 w - -", 10);
  test("WAC.009", "d6h2", "3q1rk1/p4pp1/2pb3p/3p4/6Pr/1PNQ4/P1PB1PP1/4RRK1 b - -", 10);
  test("WAC.010", "h4h7", "2br2k1/2q3rn/p2NppQ1/2p1P3/Pp5R/4P3/1P3PPP/3R2K1 w - -", 9);
  test("WAC.011", "f3c6", "r1b1kb1r/3q1ppp/pBp1pn2/8/Np3P2/5B2/PPP3PP/R2Q1RK1 w kq -", 7);
  test("WAC.012", "g4f3", "4k1r1/2p3r1/1pR1p3/3pP2p/3P2qP/P4N2/1PQ4P/5R1K b - -", 7);
  test("WAC.013", "f1f8", "5rk1/pp4p1/2n1p2p/2Npq3/2p5/6P1/P3P1BP/R4Q1K w - -", 9);
  test("WAC.014", "h3h7", "r2rb1k1/pp1q1p1p/2n1p1p1/2bp4/5P2/PP1BPR1Q/1BPN2PP/R5K1 w - -", 9);
  test("WAC.015", "b8b7", "1R6/1brk2p1/4p2p/p1P1Pp2/P7/6P1/1P4P1/2R3K1 w - -", 9);
  test("WAC.016", "e2c3", "r4rk1/ppp2ppp/2n5/2bqp3/8/P2PB3/1PP1NPPP/R2Q1RK1 w - -", 7);
  test("WAC.017", "c4e5", "1k5r/pppbn1pp/4q1r1/1P3p2/2NPp3/1QP5/P4PPP/R1B1R1K1 w - -", 7);
  test("WAC.018", "a8h8", "R7/P4k2/8/8/8/8/r7/6K1 w - -", 20);
  test("WAC.019", "c5c6", "r1b2rk1/ppbn1ppp/4p3/1QP4q/3P4/N4N2/5PPP/R1B2RK1 w - -", 7);
  test("WAC.020", "d7b5", "r2qkb1r/1ppb1ppp/p7/4p3/P1Q1P3/2P5/5PPP/R1B2KNR b kq -", 7);//USTABIL
}

static void suite_5_candidates_2020_p1_move11() {
  
  
  test("rd1-lagrave-caruana", "h2h3", "1rbqk2r/b1p2ppp/p1np1n2/Pp2p3/3PP3/1BP2N2/1P3PPP/RNBQ1RK1 w k - 1 11", 9);
  test("rd1-ding-hao", "e2e4", "rnbqr1k1/ppb2ppp/2pp1n2/4p3/NPP5/P2P1NP1/4PPBP/R1BQ1RK1 w - - 1 11", 9);
  test("rd1-giri-nepomniachtchi", "h3g4", "r1b1kb1r/1p1p1ppp/1q2p3/1p2n3/2P2Bn1/2N1P1PP/PP3P2/R2QKB1R w KQkq - 0 11", 9);
  test("rd1-grischuk-alekseenko", "c1b2", "1rbq1rk1/ppp2pp1/2n4p/2b1p3/8/2PP1NP1/P1Q1PPBP/R1B2RK1 w - - 2 11", 9);
  
  test("rd2-caruana-alekseenko", "g1f3", "rnbqr1k1/pp2bppp/5n2/2pP4/2P5/P1N5/1P2B1PP/R1BQK1NR w KQ - 3 11", 9);
  test("rd2-nepomniachtchi-grischuk", "c1f4", "r1b1kb1r/ppp2pp1/2p5/4Pn1p/8/2N2N1P/PPP2PP1/R1B2RK1 w - - 0 11", 9);
  test("rd2-hao-giri", "c1f4", "r1bqk2r/pp3pb1/2n2npp/1N1p4/3p3P/5NP1/PP2PPB1/R1BQ1RK1 w kq - 2 11", 9);
  test("rd2-lagrave-ding", "b1c3", "r4rk1/1bpqbppp/p1np1n2/1p2p3/4P3/PB1P1N1P/1PP2PP1/RNBQR1K1 w - - 1 11", 9);
  
  test("rd3-liren-caruana", "d1d2", "rn1q1rk1/pp3ppp/2p2n2/4N3/Pb1P4/2N2P2/1Pb1PKPP/R1BQ1B1R w - - 1 11", 9);
  test("rd3-giri-lagrave", "f2f4", "rnb2rk1/pp2pp1p/3q2p1/2pP4/4P3/2P5/P2Q1PPP/R3KBNR w KQ - 3 11", 9);
  test("rd3-grischuk-hao", "c2c3", "rn2k2r/ppp1q1pp/3bpn2/3p4/3P4/3B4/PPP1QPPP/RNB1K2R w KQkq - 0 11", 9);
  test("rd3-alekseenko-nepomniachtchi", "e1g1", "rn2k2r/p1q1nppp/1p2p3/2ppP3/b2P3P/P1P2N2/2P1BPP1/R1BQK2R w KQkq - 4 11", 9);
  
  //  test("rd3-", "", "", 7);
  //  test("rd3-", "", "", 7);
  //  test("rd3-", "", "", 7);
  //  test("rd3-", "", "", 7);
}

static void suite_fornax_games() {
  test("FG1", "g4f2", "7k/2p5/rp6/5p1p/3RbQnP/6PK/2qB4/3R4 b - - 4 38", 9);
  test("FG2", "d4c4", "5q1k/8/1PQp2p1/1p2p2p/3r3P/5BP1/3P4/4K3 b - - 0 40", 11);
  test("FG3", "e8d7", "r1bqk2r/pp2b2p/2np4/2pQpp2/2P1P3/5N2/PP2BPPP/R1B2RK1 b kq - 1 11", 10);
  test("FG4", "e2f3", "8/r4q1k/2P1QP1b/3p1P2/p7/6RP/P1r1K1B1/5R2 w - - 3 38", 13);
}

static int fail;
static int success;
static int total;

void searchsuite_number(int n) {
  fail = 0;
  success = 0;
  total = 0;
  stopwatch_wall_start();
  switch (n) {
    case 1:
      suite_1_fast();
      break;
    case 2:
      suite_2_solids();
      break;
    case 3:
      suite_3_bkfull();
      break;
    case 4:
      suite_4_win_at_chess();
      break;
    case 5:
      suite_5_candidates_2020_p1_move11();
      break;
    case 6:
      suite_fornax_games();
      break;
    default:
      break;
  }
  stopwatch_wall_stop_print();
  assert(fail + success == total);
  printf("SEARCH TESTS %d/%d \n", success, total);
}

void searchsuite_standard(void) {
  fail = 0;
  success = 0;
  total = 0;
  stopwatch_wall_start();
  suite_1_fast();
  suite_2_solids();
  stopwatch_wall_stop_print();
  assert(fail + success == total);
  printf("SEARCH TESTS %d/%d \n", success, total);
}

static void test(const char* name, char* should, const char* fen, uint8_t depth) {
  BoardCopiedState aux;
  Board b;
  b.copyState = &aux;
  fen_to_board(&b, fen);
  printf("\ntesting %s ...\n", name);
  boardprint_fen(b);
  printf("\n");
  
  move s = move_full_from_uci(&b, should, (int) strlen(should));
  move was = search_iterative(&b, depth);
  
  if (s != was) {
    printf("SEARCH FAIL: '%s' depth=%d (should ", name, depth);
    boardprint_move(s);
    printf(" was ");
    boardprint_move(was);
    printf(") \n");
    fail++;
  }
  else {
    success++;
  }
  total++;
}
