//
//  searchmodes.c
//  fornax2_core
//
//  Created by Anders on 15/06/2019.
//  Copyright © 2019 Avividi. All rights reserved.
//

#include <pthread.h>
#include <limits.h>
#include <stdlib.h>
#include <time.h>
#include <stdio.h>

#include "searchmodes.h"
#include "search.h"
#include "../genericutils/stopwatch.h"
#include "../genericutils/debugprint.h"


move searchmodes_best = 0;
static bool stopflag = false;

typedef struct search_args {
  const Board* board;
  uint8_t depth;
} search_args;


static void *_searchdepth_inthread(void *arg) {
#if DEBUG
  stopwatch_cpu_start();
#endif
  search_args* searchargs = (search_args*) arg;
  stopflag = false;
  searchmodes_best = search_iterative_stop(searchargs->board, searchargs->depth, &stopflag);
  
#if DEBUG
  stopwatch_cpu_stop_print();
#endif
  free(arg);
  return NULL;
}

void searchmodes_depth(const Board* board, uint8_t depth) {
  search_args* args = malloc(sizeof(search_args));
  args->board = board;
  args->depth = depth;
  
  pthread_t t1;
  pthread_create(&t1, NULL, _searchdepth_inthread, (void*) args);
}

void searchmodes_infinite(const Board* board) {
  searchmodes_depth(board, SEARCH_DEPTH_MAX);
}

void searchmodes_stop(void) {
  stopflag = true;
}

static void *timer(void *arg) {

  long time = *((long*)arg);
  long seconds = time / 1000;
  long nanoseconds = (time % 1000) * 1000 * 1000;

//  printf("seconds %ld, nanos %ld\n", seconds, nanoseconds);
  nanosleep((const struct timespec[]){{seconds, nanoseconds}}, NULL);

  searchmodes_stop();
  free(arg);
  return NULL;
}

void searchmodes_time(const Board* board, long time) {
  debug_print_l1("go time %ld\n", time);

  pthread_t t2;
  long* ttime = malloc(sizeof(*ttime));
  *ttime = time;
  pthread_create(&t2, NULL, timer, (void*) ttime);

  searchmodes_infinite(board);
}
