//
//  zobrist.c
//  fornax2_core
//
//  Created by Anders on 19/06/2019.
//  Copyright © 2019 Avividi. All rights reserved.
//

#include <stdlib.h>
#include <stdbool.h>

#include "zobrist.h"
#include "../board/definitions.h"
#include "../board/bitboard.h"
#include "../board/square.h"
#include "../boardprint.h"
#include "zobristnumbers.inc.h"

#ifndef NDEBUG
static const int size = 64 * 12 + 4 + 1 + 8;
#endif

ttkey zobrist_piece(square pos, inhabitant inhabitant, bool white) {
  int index = inhabitant * 64 + pos + 64 * 6 * white;
  assert(index < 64 * 12);
  return zobrist_numbers[index];
}

ttkey zobrist_white_move(void) {
  return zobrist_numbers[64 * 12];
}

static inline ttkey castleWhiteQ() {
  return zobrist_numbers[64 * 12 + 1];
}

static inline ttkey castleWhiteK() {
  return zobrist_numbers[64 * 12 + 2];
}

static inline ttkey castleBlackQ() {
  return zobrist_numbers[64 * 12 + 3];
}

static inline ttkey castleBlackK() {
  return zobrist_numbers[64 * 12 + 4];
}

ttkey zobrist_castle_k(bool white) {
  return zobrist_numbers[64 * 12 + 4 - white - white];
}

ttkey zobrist_castle_q(bool white) {
  return zobrist_numbers[64 * 12 + 3 - white - white];
}


static inline bool canCastleWhiteK(const Board *b) {
  return (b->isWhiteMove && b->copyState->fck)
  || (!b->isWhiteMove && b->copyState->hck);
}

static inline bool canCastleWhiteQ(const Board *b) {
  return (b->isWhiteMove && b->copyState->fcq)
  || (!b->isWhiteMove && b->copyState->hcq);
}

static inline bool canCastleBlackK(const Board *b) {
  return (b->isWhiteMove && b->copyState->hck)
  || (!b->isWhiteMove && b->copyState->fck);
}

static inline bool canCastleBlackQ(const Board *b) {
  return (b->isWhiteMove && b->copyState->hcq)
  || (!b->isWhiteMove && b->copyState->fcq);
}


ttkey zobrist_enpassant(bitboard64 passant) {
  int pos = bitboard_findLeading(passant);
  if (pos < 0) return 0;
  Square s = Square_unflatten(pos);
  
  int i = s.x;
  
  int index = 64 * 12 + 5 + i;
  assert(index < size);
  return zobrist_numbers[index];
}

ttkey zobrist_get(const Board* b) {
  ttkey hash = 0;
  
  for (int i = 0; i < 6; i++) {
    bitboard64 bitboard = b->white.pieces[i];
    int pos;
    while ((pos = bitboard_popTrailing(&bitboard)) >= 0) {
      hash ^= zobrist_piece(pos, i, 1);
    }
    bitboard = b->black.pieces[i];
    while ((pos = bitboard_popTrailing(&bitboard)) >= 0) {
      hash ^= zobrist_piece(pos, i, 0);
    }
  }
  if (b->isWhiteMove) hash ^= zobrist_white_move();
  if (canCastleWhiteK(b)) hash ^= castleWhiteK();
  if (canCastleWhiteQ(b)) hash ^= castleWhiteQ();
  if (canCastleBlackK(b)) hash ^= castleBlackK();
  if (canCastleBlackQ(b)) hash ^= castleBlackQ();
  hash ^= zobrist_enpassant(b->copyState->enPassant);
  return hash;
}


